import unittest
from datetime import date

from src.scripts.authenticity_space import (
    compare_to_openstreetmap,
    get_coordinate_from_wikidata,
    get_QID,
)


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.today = date.today().strftime("%Y-%m-%d")

    def test_it_should_return_QID(self):
        self.assertEqual(
            get_QID("Hong Kong"),
            {"id": "Q8646", "label": "Hong Kong"},
        )

    def test_it_should_return_coordinates(self):
        self.assertEqual(
            get_coordinate_from_wikidata("Q8646"),
            [["114.158611111", "22.278333333"]],
        )

    def test_it_should_return_empty_list(self):
        self.assertEqual(
            compare_to_openstreetmap(
                {"SP0023": ["Shenyang", "PL", 41.805699, 123.431472]}
            ),
            [],
        )

    def test_it_should_return_list_of_not_match(self):
        self.assertEqual(
            compare_to_openstreetmap(
                {"SP0025": ["Baiyangdian", "PL", 38.941441, 115.969465]}
            ),
            [["Baiyangdian", "PL", 38.941441, 115.969465, "SP0025"]],
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import date

from src.scripts.authenticity_institution import get_QID_inst


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.today = date.today().strftime("%Y-%m-%d")

    def test_it_should_return_QID_info(self):
        self.assertEqual(
            get_QID_inst("August First Film Studio"),
            [{"id": "Q10872932", "label": "August First Film Studio"}],
        )

    def test_it_should_return_None_for_nothing_match(self):
        self.assertEqual(
            get_QID_inst("Jin Opera Group of Liulin County"),
            None,
        )


if __name__ == "__main__":
    unittest.main()



import importlib
import logging
import os
import sys
from datetime import date
from importlib.metadata import version

import click
import pandas as pd

from src.scripts.agent_table_processing import process_agent_tables
from src.scripts.authenticity_space import get_coordinate_from_wikidata, get_QID
from src.scripts.process_Institution import process_Inst
from src.scripts.process_Person import process_Pers
from src.scripts.process_Space import process_Spac

# Creating an object
logger = logging.getLogger()
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s: - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
SPACE_GITHUB = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Space.csv"


def combine_space_tables(df_space_user, df_space_gh, space_ids_gh):
    """
    This function aims to use merging the Space tables in ReadAct and in user's directory. If any space_id in the
    user table already exists in ReadAct, delete according line(s) in the user table and then merge.
    :param df_space_user: dataframe of the Space.csv from user
    :param df_space_gh: dataframe of the Space.csv from ReadAct
    :param space_ids_gh: all the space_ids in the Space.csv in ReadAct
    :return: a processed combined agent dataframe
    """
    df_space_user_not_in_gh = df_space_user.loc[
        ~df_space_user["space_id"].isin(space_ids_gh)
    ]  # delete lines which
    # has agent_id in ReadAct, which is to say, for any agent_id already in ReadAct, use the ReadAct resource
    space_cols = [
        "space_id",
        "old_id",
        "space_type",
        "space_name",
        "language",
        "lat",
        "long",
        "wikidata_id",
        "note",
        "created",
        "created_by",
        "last_modified",
        "last_modified_by",
    ]
    df_processd = pd.merge(
        df_space_gh, df_space_user_not_in_gh, on=space_cols, how="outer"
    )
    return df_processd


def create_new_space_entry(
    df, place_dict_combined, today, place_name, combined_two_space, entity_type, path
):
    # Read Space.csv from ReadAct
    df_space_gh = pd.read_csv(SPACE_GITHUB).fillna("")
    space_ids_gh = df_space_gh["space_id"].tolist()
    space_ids_gh.sort()
    last_space_id = space_ids_gh[-1]

    if (
        combined_two_space is True
    ):  # Already read local Space.csv. Must combine two space table.
        if entity_type == "Person":
            path_space_user = path[:-10] + "Space.csv"
        elif entity_type == "Institution":
            path_space_user = path[:-15] + "Space.csv"
        df_space_user = pd.read_csv(path_space_user)
        df_space_processed = combine_space_tables(
            df_space_user, df_space_gh, space_ids_gh
        )
    else:  # the place in user's P/I table are all in ReadAct, unnecessary to check for potential local Space.csv. Pay
        # attention that if ReadActor find any new space entity then to save a new Space table might overwrite any
        # potential local Space.csv
        df_space_processed = df_space_gh

    flag = False  # Flag to show if any new entries are added
    for index, row in df.iterrows():
        if (
            row[place_name] is None
            or pd.isna(row[place_name])
            or len(str(row[place_name])) == 0
        ):
            continue
        elif row[place_name] in place_dict_combined:
            continue
        else:
            # New space introduced by SPAQRL query
            # Append new entries
            # space_id,old_id,space_type,space_name,language,lat,long,wikidata_id,note,created,created_by, last_modified,last_modified_by
            if int(last_space_id[2:]) > 9999:
                logger.error(
                    "Please inform the maintainer to update the schema of Space and modify scripts accordingly."
                )
                sys.exit()
            new_space_id = last_space_id[0:2] + str(int(last_space_id[2:]) + 1)
            new_space_name = row[place_name]
            new_language = "en"
            new_created = today
            new_created_by = "ReadActor"
            query_space = get_QID(new_space_name)
            if query_space is None:
                new_wikidata_id = ""
                new_lat = ""
                new_long = ""
                new_space_type = "L"  # L for locations (with NULL coordinates)
            else:
                coordinate = get_coordinate_from_wikidata(query_space["id"])[0]
                new_wikidata_id = query_space["id"]
                if len(coordinate) == 0:
                    new_lat = ""
                    new_long = ""
                elif len(coordinate) == 1:
                    new_long = coordinate[0]
                    new_lat = ""
                elif len(coordinate) == 2:
                    new_lat = coordinate[1]
                    new_long = coordinate[0]
                new_space_type = "PL"  # PL for place
            df_space_processed.loc[len(df_space_processed.index)] = [
                new_space_id,
                "",  # space_type
                new_space_type,
                new_space_name,
                new_language,
                new_lat,
                new_long,
                new_wikidata_id,
                "",  # note
                new_created,
                new_created_by,
                "",  # last_modified
                "",  # last_modified_by
            ]
            flag = True
            last_space_id = new_space_id
    return df_space_processed, flag


def space_dict_for_agents(df_space):
    space_dict = {}
    for index, row in df_space.iterrows():
        # consider the case that if there are identical space_names in csv file
        if row["space_name"] not in space_dict.keys():
            # key: 'space_name'
            # value: 'space_id'
            space_dict[row["space_name"]] = row["space_id"]
        else:
            # ToDo(QG): Is it on purpose that we have reduplicated space_name ?
            logger.warning(
                "There are reduplicated space_name in ReadAct. Please notice the maintainer."
            )
            continue
    space_dict[""] = ""
    space_dict[None] = ""
    space_dict[float("nan")] = ""
    return space_dict


# eager
def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    s = "version " + importlib.metadata.version("ReadActor")
    click.echo(s)
    ctx.exit()


# eager
def print_log(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    with open("ReadActor.log", "r") as f:
        data = f.read()
        click.echo(data)
    ctx.exit()


def log(level):
    fh = logging.FileHandler("ReadActor.log")
    ch = logging.StreamHandler()

    logger.setLevel(logging.DEBUG)

    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)

    ch.setLevel(level)
    ch.setFormatter(formatter)

    logger.addHandler(ch)
    logger.addHandler(fh)


# Todo(QG): double check for the support of full URI.
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v",
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Package version",
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    callback=print_log,
    expose_value=False,
    is_eager=True,
    help="Print full log output to console",
)
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    default=False,
    is_eager=True,
    help="Prompt user for confirmation to continue",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Print no log output to console other then completion message and error level events",
)
@click.option(
    "-o",
    "--output",
    is_flag=True,
    help="Do not update input table, but create a new file at <path> instead",
)
@click.option(
    "-s",
    "--summary",
    is_flag=True,
    help="Do not update input table, but summarise results in console",
)
@click.option(
    "-S",
    "--space",
    is_flag=True,
    help="Process only places (places and locations)",
)
@click.option(
    "-A",
    "--agents",
    is_flag=True,
    help="Process only agents (persons and institutions)",
)
@click.argument("path", default=".", type=str)
def cli(path, interactive, quiet, output, summary, space, agents):
    if interactive:
        click.confirm("Do you want to update the table?", default=False, abort=True)

    if quiet:
        level = logging.ERROR
    else:
        level = logging.INFO

    log(level)

    if space:
        if "Space" not in path:
            print(
                "You want to process space/locations, but your input file path doesn't contain this kind of file."
            )
            sys.exit()
    if agents:
        if "Person" not in path and "Institution" not in path:
            print(
                "You want to process person/institution, but your input file path doesn't contain this kind of file."
            )
            sys.exit()
    today = date.today().strftime("%Y-%m-%d")

    # process the dataframe (Person, Space, Institution).
    if "Space" in path:
        entity_type = "Space"
        df = pd.read_csv(path)  # index_col=0
        df = df.fillna("")  # Replace all the nan into empty string
        df = process_Spac(df)

    elif "Person" in path:
        entity_type = "Person"
        place_name = "place_of_birth"
        agent_user_path = path[:-10] + "Agent.csv"
        (
            df,
            agent_processed,
            _,
            _,
            _,
            place_dict_combined,
            combined_two_space,
        ) = process_agent_tables(entity_type, "user", path=[path, agent_user_path])

        agent_processed_sorted = agent_processed.loc[
            agent_processed["agent_id"].str[2:].astype(int).sort_values().index
        ].reset_index(drop=True)

        df = process_Pers(df, entity_type)
        df_sorted = df.loc[
            df["person_id"].str[2:].astype(int).sort_values().index
        ].reset_index(
            drop=True
        )  # Sort Person.csv by person_id
        df = df_sorted

        df_space_raw = pd.read_csv(SPACE_GITHUB)
        space_dict = space_dict_for_agents(df_space_raw)
        df = df.replace({"place_of_birth": space_dict})
        df_space_processed, flag_space_table = create_new_space_entry(
            df,
            place_dict_combined,
            today,
            place_name,
            combined_two_space,
            entity_type,
            path,
        )

    elif "Institution" in path:
        entity_type = "Institution"
        place_name = "place"
        agent_user_path = path[:-15] + "Agent.csv"
        (
            df,
            agent_processed,
            _,
            _,
            _,
            place_dict_combined,
            combined_two_space,
        ) = process_agent_tables(entity_type, "user", path=[path, agent_user_path])
        agent_processed_sorted = agent_processed.loc[
            agent_processed["agent_id"].str[2:].astype(int).sort_values().index
        ].reset_index(drop=True)
        df = process_Inst(df, entity_type)
        df_sorted = df.loc[
            df["inst_id"].str[2:].astype(int).sort_values().index
        ].reset_index(
            drop=True
        )  # Sort Institution.csv by inst_id
        df = df_sorted

        df_space_raw = pd.read_csv(SPACE_GITHUB)
        space_dict = space_dict_for_agents(df_space_raw)
        df = df.replace({"place": space_dict})
        df_space_processed, flag_space_table = create_new_space_entry(
            df,
            place_dict_combined,
            today,
            place_name,
            combined_two_space,
            entity_type,
            path,
        )

    if entity_type == "Person":
        a_id = "person_id"
        # after adding new space entry, replace space name with new space ID
        space_dict = space_dict_for_agents(df_space_processed)
        df = df.replace({"place_of_birth": space_dict})
    elif entity_type == "Institution":
        a_id = "inst_id"
        # after adding new space entry, replace space name with new space ID
        space_dict = space_dict_for_agents(df_space_processed)
        df = df.replace({"place": space_dict})

    # output to new tables
    if output:
        if entity_type == "Space":
            new_csv_path = path[:-4] + "_updated.csv"
            with open(new_csv_path, "w+") as f:
                f.write(df.to_csv(index=False))
        else:
            # write two/three updated tables to new files: agent and the other
            df_person_or_inst = df.copy(deep=True)  # a deep copy
            df_person_or_inst.drop("wikidata_id", inplace=True, axis=1)
            new_csv_path = path[:-4] + "_updated.csv"
            with open(new_csv_path, "w") as f3:
                f3.write(df_person_or_inst.to_csv(index=False))

            df_agent = agent_processed_sorted
            for i in range(len(df_agent["agent_id"])):
                for j in range(len(df[a_id])):
                    if df_agent["agent_id"][i] == df[a_id][j]:
                        if df_agent["wikidata_id"][i] != df["wikidata_id"][j]:
                            df_agent.loc[i, "wikidata_id"] = df["wikidata_id"][j]
                            df_agent.loc[i, "last_modified"] = today
                            df_agent.loc[i, "last_modified_by"] = "ReadActor"
                            logger.info("Wikidata id is updated. ")
            new_agent_user_path = agent_user_path[:-4] + "_updated.csv"
            with open(new_agent_user_path, "w") as f:
                f.write(df_agent.to_csv(index=False))

            if flag_space_table:
                if entity_type == "Person":
                    path_space = path[:-10] + "Space_updated.csv"
                elif entity_type == "Institution":
                    path_space = path[:-15] + "Space_updated.csv"
                with open(path_space, "w") as f:
                    f.write(df_space_processed.to_csv(index=False))

    # Print summary
    elif summary:
        if entity_type == "Space":
            print("\nSummary:\n", df.to_csv(index=False))
        elif entity_type == "Person" or entity_type == "Institution":
            # print two tables on screen: agent and the other
            df_person_or_inst = df.copy(deep=True)  # a deep copy
            df_person_or_inst.drop("wikidata_id", inplace=True, axis=1)
            print("\nSummary of Person/Institution:")
            print(df_person_or_inst.to_csv(index=False))

            df_agent = agent_processed_sorted
            for i in range(len(df_agent["agent_id"])):
                for j in range(len(df[a_id])):
                    if df_agent["agent_id"][i] == df[a_id][j]:
                        if df_agent["wikidata_id"][i] != df["wikidata_id"][j]:
                            df_agent.loc[i, "wikidata_id"] = df["wikidata_id"][j]
                            df_agent.loc[i, "last_modified"] = today
                            df_agent.loc[i, "last_modified_by"] = "ReadActor"
                            logger.info("Wikidata id is updated. ")
            # print("\nSummary of Agent:")
            # print(df_agent.to_csv(index=False))

            if flag_space_table:
                print("\nSummary of Space")
                print(df_space_processed.to_csv(index=False))

    else:
        if entity_type == "Space":
            with open(path, "w") as f1:
                f1.write(df.to_csv(index=False))
        else:
            # updated two tables: agent and the other

            df_person_or_inst = df.copy(deep=True)  # a deep copy
            df_person_or_inst.drop("wikidata_id", inplace=True, axis=1)

            with open(path, "w") as f3:
                f3.write(df_person_or_inst.to_csv(index=False))

            df_agent = agent_processed_sorted
            for i in range(len(df_agent["agent_id"])):
                for j in range(len(df[a_id])):
                    if df_agent["agent_id"][i] == df[a_id][j]:
                        if df_agent["wikidata_id"][i] != df["wikidata_id"][j]:
                            df_agent.loc[i, "wikidata_id"] = df["wikidata_id"][j]
                            df_agent.loc[i, "last_modified"] = today
                            df_agent.loc[i, "last_modified_by"] = "ReadActor"
                            logger.info("Wikidata id is updated. ")
            with open(agent_user_path, "w") as f:
                f.write(df_agent.to_csv(index=False))

            if flag_space_table:
                if entity_type == "Person":
                    path_space = path[:-10] + "Space.csv"
                elif entity_type == "Institution":
                    path_space = path[:-15] + "Space.csv"
                if os.path.isfile(path_space):
                    logger.warning("Your Space.csv at %s is overwritten. " % path_space)
                with open(path_space, "w") as f:
                    f.write(df_space_processed.to_csv(index=False))


if __name__ == "__main__":
    cli()



import datetime
import os
import re
import unittest
from os import listdir
from os.path import exists, isfile, join

import click
from click.testing import CliRunner

from src.scripts.readactor import cli


class TestSum(unittest.TestCase):
    def test_help_1_should_return_documentation(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["--help"])
        print(result.output.splitlines())
        assert result.exit_code == 0
        assert result.output.splitlines() == [
            "Usage: cli [OPTIONS] [PATH]",
            "",
            "Options:",
            "  -v, --version      Package version",
            "  -d, --debug        Print full log output to console",
            "  -i, --interactive  Prompt user for confirmation to continue",
            "  -q, --quiet        Print no log output to console other then completion",
            "                     message and error level events",
            "  -o, --output       Do not update input table, but create a new file at <path>",
            "                     instead",
            "  -s, --summary      Do not update input table, but summarise results in console",
            "  -S, --space        Process only places (places and locations)",
            "  -A, --agents       Process only agents (persons and institutions)",
            "  -h, --help         Show this message and exit.",
        ]

    def test_help_2_should_return_documentation(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-h"])
        print(result.output.splitlines())
        assert result.exit_code == 0
        assert result.output.splitlines() == [
            "Usage: cli [OPTIONS] [PATH]",
            "",
            "Options:",
            "  -v, --version      Package version",
            "  -d, --debug        Print full log output to console",
            "  -i, --interactive  Prompt user for confirmation to continue",
            "  -q, --quiet        Print no log output to console other then completion",
            "                     message and error level events",
            "  -o, --output       Do not update input table, but create a new file at <path>",
            "                     instead",
            "  -s, --summary      Do not update input table, but summarise results in console",
            "  -S, --space        Process only places (places and locations)",
            "  -A, --agents       Process only agents (persons and institutions)",
            "  -h, --help         Show this message and exit.",
        ]

    def test_version_1_should_return_version(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["--version"])
        assert result.exit_code == 0
        assert bool(
            re.search("version (\d+\.){2}(\d)+((-|_))*(\w)*", result.output.strip())
        )

    def test_version_2_should_return_version(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-v"])
        assert result.exit_code == 0
        assert bool(
            re.search("version (\d+\.){2}(\d)+((-|_))*(\w)*", result.output.strip())
        )

    def test_debug_1_should_return_log_text(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("ReadActor.log", "w") as f:
                f.write(
                    "2022-05-24 21:11:23 - root - WARNING: - For row 6, you should input at least a person_id even if "
                    "there is no matched wikidata_id. By SemBot.\n2022-05-24 21:25:48 - root - INFO: - Fields "
                    '"birthyear, deathyear, place_of_birth, created, created_by, last_modified, last_modified_by" '
                    "is/are overwritten.  By SemBot.\n2022-05-24 21:25:48 - urllib3.connectionpool - DEBUG: - "
                    "Starting new HTTPS connection (1): query.wikidata.org:443\n"
                )
            result = runner.invoke(cli, ["--debug"])
            pattern_log = (
                "^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01]) (\d\d:){2}\d{2} - (\w.)+\s-\s("
                "DEBUG|INFO|WARNING|ERROR|CRITICAL|NOTSET):\s-\s(.*?)+"
            )
            assert result.exit_code == 0
            assert bool(re.search(pattern_log, result.output))

    def test_debug_2_should_return_log_text(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("ReadActor.log", "w") as f:
                f.write(
                    "2022-05-24 21:11:23 - root - WARNING: - For row 6, you should input at least a person_id even if "
                    "there is no matched wikidata_id. By SemBot.\n2022-05-24 21:25:48 - root - INFO: - Fields "
                    '"birthyear, deathyear, place_of_birth, created, created_by, last_modified, last_modified_by" '
                    "is/are overwritten.  By SemBot.\n2022-05-24 21:25:48 - urllib3.connectionpool - DEBUG: - "
                    "Starting new HTTPS connection (1): query.wikidata.org:443\n"
                )
            result = runner.invoke(cli, ["-d"])
            pattern_log = (
                "^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01]) (\d\d:){2}\d{2} - (\w.)+\s-\s("
                "DEBUG|INFO|WARNING|ERROR|CRITICAL|NOTSET):\s-\s(.*?)+"
            )
            assert result.exit_code == 0
            assert bool(re.search(pattern_log, result.output))

    def test_interactive_1_should_be_aborted(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["--interactive", "path"], input="n")
        assert result.exit_code == 1
        assert "Aborted!" in result.output

    def test_interactive_2_should_be_aborted(self):
        runner = CliRunner()
        result = runner.invoke(cli, ["-i", "path"], input="n")
        assert result.exit_code == 1
        assert (
            "Aborted!" in result.output
        )  # (QG) result.output.strip().splitlines() == ['Do you want to update the
        # table? [y/N]: n', 'Aborted!']

    # def test_quiet_1_should_not_print_DEBUG(self):
    #     runner = CliRunner()
    #     with runner.isolated_filesystem():
    #         with open("test.csv", "w") as f:
    #             f.write(
    #                 "1,person_id,family_name,first_name,name_lang,sex,birthyear,deathyear,place_of_birth,wikidata_id,"
    #                 "created,created_by,last_modified,last_modified_by,note\n2,AG2000,Zhang,San,en,male,1999,,Berlin,"
    #                 ",2021-12-22,QG,,,skip\n3,AG0001,鲁,迅,zh,male,1881,1936,Shaoxing,Q23114,2021-12-22,QG,,,"
    #             )
    #         result = runner.invoke(cli, ["--quiet", "test.csv"])
    #         assert result.exit_code == 0
    #         assert (
    #             "DEBUG" not in result.output
    #         )  # (QG) Can be rewritten into more detailed assertion like splitlines
    #         # and then match exact output
    #         assert (
    #             "INFO" not in result.output
    #         )  # (QG) Can be rewritten into more detailed assertion like splitlines
    #         # and then match exact output

    # def test_quiet_2_should_not_print_DEBUG(self):
    #     runner = CliRunner()
    #     with runner.isolated_filesystem():
    #         with open("test.csv", "w") as f:
    #             f.write(
    #                 "1,person_id,family_name,first_name,name_lang,sex,birthyear,deathyear,place_of_birth,wikidata_id,"
    #                 "created,created_by,last_modified,last_modified_by,note\n2,AG2000,Zhang,San,en,male,1999,,Berlin,"
    #                 ",2021-12-22,QG,,,skip\n3,AG0001,鲁,迅,zh,male,1881,1936,Shaoxing,Q23114,2021-12-22,QG,,,"
    #             )
    #         result = runner.invoke(cli, ["-q", "test.csv"])
    #         assert result.exit_code == 0
    #         assert (
    #             "DEBUG" not in result.output
    #         )  # (QG) Can be rewritten into more detailed assertion like splitlines
    #         # and then match exact output
    #         assert (
    #             "INFO" not in result.output
    #         )  # (QG) Can be rewritten into more detailed assertion like splitlines
    #         # and then match exact output

    # def test_output_1_should_check_new_file(self):
    #     runner = CliRunner()
    #     with runner.isolated_filesystem():
    #         with open("Person.csv", "w") as f1:
    #             f1.write(
    #                 "1,person_id,family_name,first_name,name_lang,sex,birthyear,deathyear,place_of_birth,wikidata_id,"
    #                 "created,created_by,last_modified,last_modified_by,note\n2,AG2000,Zhang,San,en,male,1999,,Berlin,"
    #                 ",2021-12-22,QG,,,skip\n3,AG0001,鲁,迅,zh,male,1881,1936,Shaoxing,Q23114,2021-12-22,QG,,,"
    #             )
    #         #     data = {'1':  ['2', '3',],
    #         #             'person_id': ['AG2000', 'AG0001',],
    #         #             'family_name': ['Zhang', '鲁',],
    #         #            'first_name': ['San', '讯',],
    #         #            'name_lang': ['en', 'zh',],
    #         #            'sex': ['male', 'male',],
    #         #            'birthyear': ['1999', '1881',],
    #         #            'deathyear': ['', '1936',],
    #         #            'place_of_birth': ['Berlin', 'Shaoxing',],
    #         #            'wikidata_id': ['', 'Q23114',],
    #         #            'created': ['2021-12-22', '2021-12-22',],
    #         #            'created_by': ['QG', 'QG',],
    #         #            'last_modified': ['', '',],
    #         #            'last_modified_by': ['', '',],
    #         #            'note': ['skip', '',]
    #         # }
    #         #     df = pd.DataFrame(data)
    #         #     df.to_csv(f1, index=False)
    #         result = runner.invoke(cli, ["--output", "Person.csv"])
    #         assert result.exit_code == 0
    #         assert exists(
    #             "Person_updated.csv"
    #         )  # (QG) check if new file is created or not
    #         if result.output:
    #             click.echo(result.output)
    #         # to get the current working directory
    #         directory = os.getcwd()
    #         print(directory)
    #         onlyfiles = [f for f in listdir(directory) if isfile(join(directory, f))]
    #         print("onlyfiles: ", onlyfiles)
    #         with open("Person_updated.csv", "r") as f2:
    #             update = f2.read()
    #             print("update: \n", update)
    #         assert "ReadActor" in update  # (QG) make sure it is not an empty file

    # # TODO(QG): it might be more meaningful if this test can be rewritten to raise error and catch the error
    # def test_output_2_should_check_new_file(self):
    #     runner = CliRunner()
    #     with runner.isolated_filesystem():
    #         with open("Person.csv", "w") as f:
    #             f.write(
    #                 "1,person_id,family_name,first_name,name_lang,sex,birthyear,deathyear,place_of_birth,wikidata_id,"
    #                 "created,created_by,last_modified,last_modified_by,note\n2,AG2000,Zhang,San,en,male,1999,,Berlin,"
    #                 ",2021-12-22,QG,,,skip\n3,AG0001,鲁,迅,zh,male,1881,1936,Shaoxing,Q23114,2021-12-22,QG,,,"
    #             )
    #         result = runner.invoke(cli, ["-o", "Person.csv"])
    #         directory = os.getcwd()
    #         with open("Person_updated.csv", "r") as f2:
    #             update = f2.read()
    #         assert "ReadActor" in update  # (QG) make sure it is not an empty file

    # def test_summary_1_should_check_new_file(self):
    #     runner = CliRunner()
    #     with runner.isolated_filesystem():
    #         with open("Person.csv", "w") as f1:
    #             file_1 = "1,person_id,family_name,first_name,name_lang,sex,birthyear,deathyear,place_of_birth,wikidata_id,created,created_by,last_modified,last_modified_by,note\n2,AG2000,Zhang,San,en,male,1999,,Berlin,,,2021-12-22,QG,,,skip\n3,AG0001,鲁,迅,zh,male,1881,1936,Shaoxing,Q23114,2021-12-22,QG,,,"
    #             f1.write(
    #                 file_1)
    #         with open("Person.csv", "r") as f2:
    #             file_2 = f2.read()
    #             print("file_content2:\n", file_2)
    #         result = runner.invoke(cli, ["--summary", "-q", "Person.csv"])
    #         print("result.output: \n", result.output)
    #         y = result.output.strip().splitlines()
    #         print("y:\n", y)
    #         assert result.exit_code == 0
    #         assert not exists("Person_updated.csv")
    #         assert file_1 == file_2
    #         assert result.output.strip().splitlines() == ['-------------', 'For row  2  :', "['AG2000', 'Zhang', 'San', 'en', 'male', 1999, '', 'Berlin', '', '', '2021-12-22', 'QG', '', '', 'skip']", '-------------', 'For row  3  :', "['AG0001', '鲁', '迅', 'zh', 'male', 1881, 1930, 'Shaoxing', 'Q23114', '2021-12-22', 'QG', '', '', '', '']", '1,person_id,family_name,first_name,name_lang,sex,birthyear,deathyear,place_of_birth,wikidata_id,created,created_by,last_modified,last_modified_by,note', 'AG2000,Zhang,San,en,male,1999,,Berlin,,,2021-12-22,QG,,,skip', 'AG0001,鲁,迅,zh,male,1881,1936.0,aoxing,Q23114,2021-12-22,QG,,,,']


if __name__ == "__main__":
    unittest.main()



import logging
import sys
from datetime import date

from src.scripts.agent_table_processing import process_agent_tables
from src.scripts.authenticity_person import (
    order_name_by_language,
    sparql_by_name,
    sparql_with_Qid,
)

logger = logging.getLogger(__name__)


# def modify_note_lastModified_lastModifiedBy(row, message, today):
#     row["note"] += " " + message
#     row["last_modified"] = today
#     row["last_modified_by"] = "ReadActor"
#     return row


def process_Pers(df, entity_type):
    # Process the local Agent table
    (
        df_person_gh,
        agent_processed,
        person_ids_gh,
        last_person_id,
        all_wikidata_ids,
        _,
        _,
    ) = process_agent_tables(entity_type, "ReadAct", path=[])

    # Process local table row by row
    for index, row in df.iterrows():
        print("-------------\nFor row ", index + 2, " :")
        print(row.tolist())
        row, last_person_id = check_each_row_Person(
            index, row, df_person_gh, person_ids_gh, last_person_id, all_wikidata_ids
        )
        # make the format of birth and death year valid
        row = format_year_Person(row)
        df.loc[index] = row
    return df


def check_each_row_Person(
    index, row, df_person_gh, person_ids_gh, last_person_id, all_wikidata_ids
):
    today = date.today().strftime("%Y-%m-%d")
    if row["note"] == "skip" or row["note"] == "Skip":
        return row, last_person_id
    else:
        if isinstance(row["person_id"], str) and len(row["person_id"]) > 0:
            if row["person_id"] in person_ids_gh:
                return (
                    __compare_wikidata_ids_Person(index, row, df_person_gh),
                    last_person_id,
                )
            else:
                if (isinstance(row["wikidata_id"], str) is True) and (
                    len(row["wikidata_id"]) > 0
                ):
                    if row["wikidata_id"] in all_wikidata_ids:
                        row["note"] = (
                            "Error: `wikidata_id` already exists in GitHub data but the person_id does not match. "
                            "Please check. By ReadActor."
                        )
                        error_msg = (
                            "For row "
                            + str(int(index))
                            + " :"
                            + "`wikidata_id` already exists in "
                            "GitHub data "
                            "but the `person_id` does not match. Please check."
                        )
                        logger.error(error_msg)
                        sys.exit()
                    else:
                        wikidata_id_usr = row["wikidata_id"]
                        person_dict = sparql_with_Qid(wikidata_id_usr)
                        note_flag = False
                        modified_fields = []
                        if (
                            "gender" in person_dict
                            and row["sex"] != person_dict["gender"]
                        ):
                            row["sex"] = person_dict["gender"]
                            modified_fields.append("sex")
                            note_flag = True
                        if (
                            "birthyear" in person_dict
                            and row["birthyear"] != person_dict["birthyear"]
                        ):
                            row["birthyear"] = person_dict["birthyear"]
                            modified_fields.append("birthyear")
                            note_flag = True
                        if (
                            "deathyear" in person_dict
                            and row["deathyear"] != person_dict["deathyear"]
                        ):
                            row["deathyear"] = person_dict["deathyear"]
                            modified_fields.append("deathyear")
                            note_flag = True
                        if (
                            "birthplace" in person_dict
                            and row["place_of_birth"] != person_dict["birthplace"]
                        ):
                            row["place_of_birth"] = person_dict["birthplace"]
                            modified_fields.append("place_of_birth")
                            note_flag = True
                        if note_flag:
                            message = (
                                "Fields -"
                                + ", ".join(modified_fields)
                                + "- is/are overwritten.  By ReadActor."
                            )
                            if isinstance(row["note"], str):
                                row["note"] = row["note"] + " " + message
                            else:
                                row["note"] = message
                            logger.info(message)
                            row["last_modified"] = today
                            row["last_modified_by"] = "ReadActor"
                            return row, last_person_id
                        else:
                            logger.info("Row %s is checked. Pass " % index)
                            return row, last_person_id
                else:  # user provided "person_id" but not "wikidata_id"
                    names = order_name_by_language(row)
                    person = sparql_by_name(names, row["language"], 2)
                    if len(person) > 0:
                        wikidata_id_usr = next(iter(person))
                        if wikidata_id_usr in all_wikidata_ids:
                            row["note"] = (
                                "Error: `wikidata_id` queried by family_name, first_name, language already exists in "
                                ""
                                "ReadAct data, but your provided person_id does not match. Please check your data "
                                "carefully. If you are 100% sure that your input is correct, then it is likely that "
                                "this person has an identical name with a person in Wikidata database. Please put "
                                '"skip" in "note" column for this row and run this program again. By ReadActor.'
                            )
                            error_msg = (
                                "For row "
                                + str(int(index))
                                + " :"
                                + " `wikidata_id` queried by "
                                "family_name, "
                                "first_name, language already exists in ReadAct data, but your provided "
                                "person_id does not match. Please check your data carefully. If you are 100% "
                                "sure that your input is correct, then it is likely that this person has an "
                                'identical name with a person in Wikidata database. Please put "skip" in "note" '
                                "column for this row and run this program again. By ReadActor."
                            )
                            logger.error(error_msg)
                            sys.exit()
                        else:
                            row["wikidata_id"] = wikidata_id_usr
                            person_dict = sparql_with_Qid(wikidata_id_usr)
                            note_flag = False
                            modified_fields = ["wikidata_id"]
                            if (
                                "gender" in person_dict
                                and row["sex"] != person_dict["gender"]
                            ):
                                row["sex"] = person_dict["gender"]
                                modified_fields.append("sex")
                                note_flag = True
                            if (
                                "birthyear" in person_dict
                                and row["birthyear"] != person_dict["birthyear"]
                            ):
                                row["birthyear"] = person_dict["birthyear"]
                                modified_fields.append("birthyear")
                                note_flag = True
                            if (
                                "deathyear" in person_dict
                                and row["deathyear"] != person_dict["deathyear"]
                            ):
                                row["deathyear"] = person_dict["deathyear"]
                                modified_fields.append("deathyear")
                                note_flag = True
                            if (
                                "birthplace" in person_dict
                                and row["place_of_birth"] != person_dict["birthplace"]
                            ):
                                row["place_of_birth"] = person_dict["birthplace"]
                                modified_fields.append("place_of_birth")
                                note_flag = True
                            if note_flag:
                                if isinstance(row["note"], str):
                                    row["note"] = (
                                        row["note"]
                                        + " Fields "
                                        + ", ".join(modified_fields)
                                        + " in this table is/are updated.  By ReadActor."
                                    )
                                else:
                                    row["note"] = (
                                        "Fields "
                                        + ", ".join(modified_fields)
                                        + " is/are updated.  By ReadActor."
                                    )
                                logger.warning(
                                    "You should look row %s up in Wikidata again. If it does not "
                                    "match with this modification, you should retrieve the old data for "
                                    "this row and put 'skip' in 'note'." % index
                                )
                                row["last_modified"] = today
                                row["last_modified_by"] = "ReadActor"
                                return row, last_person_id
                            else:
                                logger.warning(
                                    "You should look the person in row %s up in Wikidata and input the "
                                    "`wikidata_id` in your table in the future." % index
                                )
                                if isinstance(row["note"], str):
                                    row["note"] = (
                                        row["note"]
                                        + " Field `wikidata_id` in this table is updated.  By ReadActor."
                                    )
                                else:
                                    row[
                                        "note"
                                    ] = "Field `wikidata_id` in this table is updated.  By ReadActor."
                                logger.info(
                                    "'Field `wikidata_id` in row %s of this table is updated.  By ReadActor.'"
                                    % index
                                )
                                row["last_modified"] = today
                                row["last_modified_by"] = "ReadActor"
                                return row, last_person_id
                    else:
                        if isinstance(row["note"], str):
                            row["note"] = (
                                row["note"] + " No match in Wikidata.  By ReadActor."
                            )
                        else:
                            row["note"] = "No match in Wikidata.  By ReadActor."
                        # Todo: "note" is changed, does it count as modified?
                        logger.info("Row %s in this table is checked. Pass." % index)
                        return row, last_person_id
        else:  # No user-provided `person_id`
            print("row: ", row)
            logger.error(
                'Row %s has no "person_id". You should give an unique id to each row.'
                % index
            )
            sys.exit()


def __compare_two_rows_Person(row, row_gh):
    """
    This function will be triggered when `person_id` and `wikidata_id` are the same.
    It will compare the rest fields of two rows from two dataframes seperately.

    :param row: one row from the user-uploaded Person.csv
    :param row_gh: the row in Person.csv on GitHub which has the same `person_id` and `wikidata_id` with the other
    parameter
    :return: True if all are matching, otherwise False
    """
    fields_to_be_compared = [
        "family_name",
        "first_name",
        "language",
        "sex",
        "birthyear",
        "deathyear",
        "place_of_birth",
        "created",
        "created_by",
        "last_modified",
        "last_modified_by",
        "note",
    ]
    for i in fields_to_be_compared:
        if row[i] != row_gh[i]:
            return False
    return True


def __overwrite_Person(row, row_gh):
    """
    This function will overwrite all the fields except `person_id` and `wikidata_id`.
    :param row: one row from the user-uploaded Person.csv
    :param row_gh: the row in Person.csv on GitHub which has the same `person_id` and `wikidata_id` with the other
    parameter
    :return: row which is modified
    """
    fields_to_be_overwritten = [
        "family_name",
        "first_name",
        "language",
        "sex",
        "birthyear",
        "deathyear",
        "place_of_birth",
        "created",
        "created_by",
        "last_modified",
        "last_modified_by",
        "note",
    ]
    note_flag = False
    modified_fields = []
    for i in fields_to_be_overwritten:
        if row[i] != row_gh[i]:
            row[i] = row_gh[i]
            modified_fields.append(i)
            note_flag = True
    if note_flag:
        message = (
            "Fields -"
            + ", ".join(modified_fields)
            + "- is/are overwritten.  By ReadActor."
        )
        if isinstance(row["note"], str):
            row["note"] = row["note"] + " " + message
        else:
            row["note"] = message
        logger.info(message)
    return row


def __compare_wikidata_ids_Person(index, row, df_person_gh):
    wikidata_id_usr = row["wikidata_id"]
    row_gh_index = df_person_gh.index[
        (df_person_gh["person_id"] == row["person_id"])
        & (df_person_gh["language"] == row["language"])
    ].tolist()[0]
    row_GH = df_person_gh.iloc[row_gh_index]
    wikidata_id_gh = row_GH["wikidata_id"]
    if wikidata_id_gh == wikidata_id_usr:
        res = __compare_two_rows_Person(row, row_GH)
        if not res:
            return __overwrite_Person(row, row_GH)
        logger.info("Row %s is checked. Pass " % index)
        return row
    else:
        row[
            "note"
        ] = "Error: `wikidata_id` is not matching with GitHub data. Please check. By ReadActor."
        error_msg = (
            "For row "
            + str(int(index))
            + " : `wikidata_id` does not match GitHub data. Please check. By "
            "ReadActor."
        )
        logger.error(error_msg)
        sys.exit()


def __check_person_id_size(row, last_id_in_gh):
    if int(last_id_in_gh[2:]) >= 9999:
        logger.warning(
            "It is better to update all person_id in the database. By ReadActor."
        )
        if isinstance(row["note"], str):
            row["note"] = (
                row["note"]
                + " Warning: It is better to update all person_id in the database. By ReadActor."
            )
        else:
            row[
                "note"
            ] = "Warning: It is better to update all person_id in the database. By ReadActor."


def format_year_Person(row):
    for x in ["birthyear", "deathyear"]:
        if isinstance(row[x], float):
            row[x] = int(row[x])
        if "-" in str(row[x]):
            if isinstance(row[x][1:], float):
                row[x] = "-" + str(int(row[x][1:]))
            if isinstance(row[x][1:], int):
                year = str(int(row[x][1:]))
                if len(year) < 3:
                    pad = 3 - len(year)  # pad = 2,1
                    if pad == 1:
                        row["start"] = "-" + "0" + year
                    elif pad == 2:
                        row["start"] = "-" + "00" + year
        elif isinstance(row[x], int):
            year = str(int(row[x]))
            if len(year) < 4:
                pad = 4 - len(year)  # pad = 3,2,1
                if pad == 1:
                    row[x] = "0" + year
                elif pad == 2:
                    row[x] = "00" + year
                elif pad == 3:
                    row[x] = "000" + year
    return row


if __name__ == "__main__":
    pass



"""
This is a python script to cross-reference Person entities in ReadAct with wikidata.
Strategy:
- Read ReadAct CSV files, get names (or Wikidata link) as lookups
- Get the QIDs: look up with name (or look up with Wikipedia link (if available))
- Use SPARQL to retrieve properties from the found QIDs
- Compare wikidata item properties with data in the CSV table

"""
import json
import time
from itertools import islice

import pandas as pd
import requests

from src.scripts.authenticity_space import read_space_csv

URL = "https://query.wikidata.org/sparql"
QUERY = """
        SELECT ?person ?personLabel ?ybirth ?ydeath ?birthplaceLabel ?genderLabel
        WHERE {{ 
        {{?person wdt:P31 wd:Q5 ;
                rdfs:label "{}"@{} . }} UNION {{?person wdt:P31 wd:Q5 ;
                skos:altLabel "{}"@{} . }}
        OPTIONAL {{ ?person  wdt:P21  ?gender . }}
        OPTIONAL {{ ?person  wdt:P569  ?birth . BIND(year(?birth) as ?ybirth) }}
        OPTIONAL {{ ?person  wdt:P570  ?death . BIND(year(?death) as ?ydeath) }}
        OPTIONAL {{ ?person wdt:P19  ?birthplace . }}
        OPTIONAL {{ ?person skos:altLabel ?altLabel . }}
        
        SERVICE wikibase:label {{ bd:serviceParam wikibase:language  "[AUTO_LANGUAGE], en"}}
        }}
        GROUP BY ?person ?personLabel ?ybirth ?ydeath ?birthplaceLabel ?genderLabel
        LIMIT 250
        """

QUERY_WITH_QID = """
SELECT ?person ?personLabel ?ybirth ?ydeath ?birthplaceLabel ?genderLabel
WHERE {{ 
  values ?person {{wd:{} }}
        OPTIONAL {{ ?person  wdt:P21  ?gender . }}
        OPTIONAL {{ ?person  wdt:P569  ?birth . BIND(year(?birth) as ?ybirth) }}
        OPTIONAL {{ ?person  wdt:P570  ?death . BIND(year(?death) as ?ydeath) }}
        OPTIONAL {{ ?person wdt:P19  ?birthplace . }}
        OPTIONAL {{ ?person skos:altLabel ?altLabel . }}
        
        SERVICE wikibase:label {{ bd:serviceParam wikibase:language  "[AUTO_LANGUAGE], en"}}
        }}
LIMIT 1
"""


#################################################################
################## Approach 1 : look up with name ##################
def read_person_csv(
    person_url="https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Person.csv",
):
    """
    A function to read "Person.csv", preprocess CSV.
    :param person_url
    :return: a dictionary
    """
    df = pd.read_csv(person_url, error_bad_lines=False).fillna("")
    person_dict = {}
    place_dict = read_space_csv()
    for index, row in df.iterrows():
        id = row["person_id"]
        print("-----\n", index, id)
        # a dictionary to collect final q_id for each unique person id
        if id not in person_dict:
            person_dict[id] = dict()

        if row["language"] not in person_dict[row["person_id"]]:
            # name_ordered is a list of a single name or multiple names
            name_ordered = order_name_by_language(row)

        # sex or gender type in Wikidata for human: male, female, non-binary, intersex, transgender female,
        # transgender male, agender.
        if row["sex"] not in [
            "male",
            "female",
            "non-binary",
            "intersex",
            "transgender female",
            "transgender ",
            "male",
            "agender",
        ]:
            row["sex"] = ""
        else:
            row["sex"] = row["sex"].strip()

        # birth_years and death_years are two lists of a single year or multiple years
        birth_years = __clean_birth_death_year_format(row["birthyear"])
        death_years = __clean_birth_death_year_format(row["deathyear"])

        if type(row["alt_name"]) != str:
            row["alt_name"] = ""
        else:
            row["alt_name"] = row["alt_name"].strip()

        # Replace space_id with the name of space
        if row["place_of_birth"] in place_dict:
            row["place_of_birth"] = place_dict[row["place_of_birth"]][0]
        else:
            print(
                "Please check why the place of birth is not in the dictionary of space."
            )

        person_dict[id][row["language"]] = [
            name_ordered,
            row["sex"],
            birth_years,
            death_years,
            row["alt_name"],
            row["place_of_birth"],
        ]
    return person_dict


def order_name_by_language(row):
    if type(row["family_name"]) != str:
        name_ordered = [row["first_name"]]
    elif type(row["first_name"]) != str:
        name_ordered = [row["family_name"]]
    elif row["language"] == "zh":
        name_ordered = [row["family_name"] + row["first_name"]]  # 毛泽东
    else:
        # Make it a list with two types of order to suit different languages
        # Since the non-Chinese names are the minority, the influence on processing speed can be tolerant
        # For example:
        # pinyin name: family name + " " + first name
        # Korean name: family name + " " + first name
        # Russian name: first name + " " + family name
        name_ordered = [
            row["first_name"] + " " + row["family_name"],
            row["family_name"] + " " + row["first_name"],
        ]
    return name_ordered


def __clean_birth_death_year_format(default_year):
    return default_year
    # char_to_remove = ["[", "]", "?", "~"]
    # cleaned_year = default_year
    # for c in char_to_remove:
    #     cleaned_year = cleaned_year.replace(c, "")
    #
    # if cleaned_year.isalpha():  # "XXXX" contain 0 information
    #     years = []
    # elif "." in cleaned_year:
    #     cleaned_year = cleaned_year.split("..")
    #     years = list(range(int(cleaned_year[0]), int(cleaned_year[1]) + 1))
    # elif "-" in cleaned_year:  # For BCE year
    #     cleaned_year = int(cleaned_year.replace("-", ""))
    #     years = [cleaned_year + 1, cleaned_year, cleaned_year - 1]
    # elif any([i.isalpha() for i in cleaned_year]):
    #     cleaned_year = [
    #         cleaned_year.replace("X", "0").replace("x", "0"),
    #         cleaned_year.replace("X", "9").replace("x", "0"),
    #     ]
    #     # Note(QG): Maybe consider to tickle the weight at this step already? since range(1000,2000)
    #     # covers 1000 years and it does not offer really useful information
    #     years = list(range(int(cleaned_year[0]), int(cleaned_year[1]) + 1))
    # elif "," in cleaned_year:
    #     cleaned_year = cleaned_year.split(",")
    #     years = list(range(int(cleaned_year[0]), int(cleaned_year[1]) + 1))
    # else:
    #     years = [int(cleaned_year)]
    # return years


def get_person_weight(person_dict, sleep=2):
    person_weight_dict = {}
    for person_id, value in person_dict.items():
        languages = value.keys()
        l = []
        person_weight_dict[person_id] = []
        for lang in languages:
            v = value[lang]
            # Use the ordered_name list as lookups
            lookup_names = v[0]

            # If there are alt_name, add it into the list of names
            if len(value[lang][4]) != 0:
                lookup_names.append(v[4])

            if lookup_names != ["anonymous"] and lookup_names != ["无名"]:
                person = sparql_by_name(lookup_names, lang, sleep)
            else:
                continue

            if len(person) == 0:
                # There is no match
                pass

            else:
                weight = 0
                weights = []
                Qids = []
                wiki = []
                for Q_id, p in person.items():
                    # all the matched fields will add weight 1 to the total weight for this Q_id
                    if "gender" in p:
                        if p["gender"] == v[1]:
                            weight += 1
                    elif "birthyear" in p:
                        if p["birthyear"] in v[2]:
                            weight += 1
                    elif "deathyear" in p:
                        if p["deathyear"] in v[3]:
                            weight += 1
                    elif "birthplace" in p:
                        if p["birthplace"] == v[5]:
                            weight += 1

                    weights.append(weight)
                    Qids.append(Q_id)
                    wiki.append(p)
                    weight = 0

                l.append(lang)
                l.append(weights)
                l.append(Qids)
                l.append(wiki)

            if len(l) > 0:
                person_weight_dict[person_id].append(l)
            l = []
    return person_weight_dict


def sparql_by_name(lookup_names, lang, sleep=2):
    if len(lookup_names) == 0:
        return None
    person = (
        {}
    )  # To collect entities which is found for the same person with different names
    with requests.Session() as s:
        for lookup in lookup_names:
            # print("++++ For this name: \n", lookup)
            response = s.get(
                URL,
                params={
                    "format": "json",
                    "query": QUERY.format(lookup.strip(), lang, lookup, lang),
                },
            )
            if response.status_code == 200:  # a successful response
                results = response.json().get("results", {}).get("bindings")
                if len(results) == 0:
                    # Didn't find the entity with this name on Wikidata
                    continue
                else:
                    for r in results:
                        person_wiki = {}
                        # If this entity is not recorded in the person dictionary yet:
                        if r["person"]["value"][31:] not in person:
                            if "person" in r:
                                person_wiki["Q-id"] = r["person"]["value"][
                                    31:
                                ]  # for example, 'Q558744'
                            if "personLabel" in r:
                                person_wiki["name"] = r["personLabel"]["value"]
                            if "genderLabel" in r:
                                person_wiki["gender"] = r["genderLabel"]["value"]
                            if "ybirth" in r:
                                person_wiki["birthyear"] = r["ybirth"]["value"]
                            if "ydeath" in r:
                                person_wiki["deathyear"] = r["ydeath"]["value"]
                            if "birthplaceLabel" in r:
                                person_wiki["birthplace"] = r["birthplaceLabel"][
                                    "value"
                                ]
                            person[person_wiki["Q-id"]] = person_wiki
            time.sleep(sleep)
    return person


def compare_weights(person_weight_dict):
    no_match_person = []
    match_person = {}
    for id, p in person_weight_dict.items():
        # No match at all
        if len(p) < 1:
            no_match_person.append(id)
        # At least one match for one language, no match for the rest language
        elif len(p) == 1:
            weights = p[0][1]
            max_weight = max(
                weights
            )  # for multiple identical weights to be the max weight, return the first max weight
            index = weights.index(max_weight)
            Qid = p[0][2][index]
            wi = p[0][3][index]
            match_person[id] = [Qid, wi]
        # There are at least one match for each language
        else:
            m = 0
            q = ""
            w = ""
            for item in p:
                if len(item[1]) == 1:
                    max_weight = item[1][0]
                    Qid = item[2][0]
                    wiki = item[3][0]
                else:
                    weights = item[1]
                    max_weight = max(
                        weights
                    )  # for multiple identical weights to be the max weight, return the first
                    index = weights.index(max_weight)
                    Qid = item[2][index]
                    wiki = item[3][index]
                if max_weight > m:
                    m = max_weight
                    q = Qid
                    w = wiki
                match_person[id] = [q, w]
    return no_match_person, match_person


def chunks(person_dict, SIZE=30):
    it = iter(person_dict)
    for i in range(0, len(person_dict), 2):
        yield {k: person_dict[k] for k in islice(it, SIZE)}


########################################################################
################## Approach 2 : look up with Wikipedia link ##################
def get_matched_by_wikipedia_url(
    person_url="https://raw.githubusercontent.com/readchina/ReadAct/master/csv/data/Person.csv",
):
    df = pd.read_csv(person_url, error_bad_lines=False)
    person_matched_by_wikipedia = {}

    count = 0
    for index, row in df.iterrows():
        id = row["person_id"]
        print(index, id)
        if id not in person_matched_by_wikipedia:
            Qid = get_Qid_from_wikipedia_url(row)
            # print("Qid: ", Qid)
            if Qid is not None:
                count += 1
                if count == 10:
                    time.sleep(90)
                    count = 0
                wiki = sparql_with_Qid(Qid)
                if len(wiki) > 0:
                    person_matched_by_wikipedia[id] = [Qid, wiki]
    return person_matched_by_wikipedia


def get_Qid_from_wikipedia_url(row):
    # link: the wikipedia link in Person.csv
    if isinstance(row["source_1"], str) and ".wikipedia.org/wiki/" in row["source_1"]:
        link = row["source_1"]
    elif isinstance(row["source_2"], str) and ".wikipedia.org/wiki/" in row["source_2"]:
        link = row["source_2"]
    else:
        return None
    if len(link) > 30:
        language = link[8:10]
        name = link[30:]
        # Use MediaWiki API to query
        url = (
            "https://"
            + language
            + ".wikipedia.org/w/api.php?action=query&prop=pageprops&titles="
            + name
            + "&format=json"
        )
        response = requests.get(url).json()
        if "pageprops" in list(response["query"]["pages"].values())[0]:
            pageprops = list(response["query"]["pages"].values())[0]["pageprops"]
            if "wikibase_item" in pageprops:
                Qid = list(response["query"]["pages"].values())[0]["pageprops"][
                    "wikibase_item"
                ]
                return Qid


def sparql_with_Qid(Qid):
    wiki_dict = {}
    with requests.Session() as s:
        response = s.get(
            URL, params={"format": "json", "query": QUERY_WITH_QID.format(Qid)}
        )
        if response.status_code == 200:  # a successful response
            results = response.json().get("results", {}).get("bindings")
            if len(results) == 0:
                print(
                    "Didn't find the entity with this Q-identifier \"",
                    Qid,
                    '" on Wikidata',
                )
                return None
            else:
                for r in results:
                    if r is not None:
                        wiki_dict["Q-id"] = Qid
                        if "personLabel" in r:
                            wiki_dict["name"] = r["personLabel"]["value"]
                        if "genderLabel" in r:
                            wiki_dict["gender"] = r["genderLabel"]["value"]
                        if "ybirth" in r:
                            wiki_dict["birthyear"] = r["ybirth"]["value"]
                        if "ydeath" in r:
                            wiki_dict["deathyear"] = r["ydeath"]["value"]
                        if "birthplaceLabel" in r:
                            wiki_dict["birthplace"] = r["birthplaceLabel"]["value"]
    return wiki_dict


if __name__ == "__main__":

    person_url = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Person.csv"

    #################################################################
    ################## Approach 1 : look up with name ##################
    person_dict = read_person_csv(
        "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Person.csv"
    )

    # Break the entire dictionary into several chunks.
    # So that we can add break sessions in between to avoid exceed the limitation of SPARQL query
    no_match_by_name = []
    matched_by_name = {}
    for chunk in chunks(person_dict, 30):  # the digit here controls the batch size
        if len(chunk) > 0:
            print("chunk: \n", chunk)
            person_weight_dict = get_person_weight(chunk, 2)
            no_match, person_match_dict = compare_weights(person_weight_dict)
            if len(no_match) > 0:
                no_match_by_name = [*no_match_by_name, *no_match]
            if len(person_match_dict) > 0:
                matched_by_name = {**matched_by_name, **person_match_dict}

            print("\n----------\n")
            print("no match: ", no_match)
            print("person_match_dict: ", person_match_dict)

            print("\n===========================\n")
            print("Current final no match: ", no_match_by_name)
            print("Current final person_match_dict: ", matched_by_name)
            print("\n I am taking a break XD \n")

            time.sleep(
                90
            )  # for every a few person entries, let this script take a break of 90 seconds

    print("I finished all the iteration.")
    print("\n===========================\n")
    print("no match: ", no_match_by_name)
    print("person_match_dict: ", matched_by_name)

    with open("../results/no_match_by_querying_name.json", "w") as f:
        json.dump(no_match_by_name, f)

    with open("../results/matched_by_name.json", "w") as f:
        json.dump(matched_by_name, f)

    ########################################################################
    ################## Approach 2 : look up with wikidata link ##################
    # matched_by_wikipedia = get_matched_by_wikipedia_url("https://raw.githubusercontent.com/readchina/ReadAct/master/csv/data/Person.csv")
    #
    # print('===========\n', matched_by_wikipedia)
    # with open('../results/matched_by_wikipedia.json', 'w') as f:
    #     json.dump(matched_by_wikipedia, f)

    #################################################################
    ################## Comparison ##################
    with open("../results/matched_by_name.json", "r") as f:
        name = json.load(f)
    with open("../results/matched_by_wikipedia.json", "r") as f:
        wikipedia = json.load(f)

    # Rearranged a dictionay to contain person information for later comparison
    person = read_person_csv(person_url)
    new_dict = {}
    for key in person.keys():
        names, gender, birthyear, deathyear, birthplace = [], [], [], [], []
        new_dict[key] = {}
        langs = [*person[key]]
        for lang in langs:
            names.extend(person[key][lang][0])
            if len(person[key][lang][4]) > 0:
                names.append(person[key][lang][4])
            gender.append(person[key][lang][1])
            birthyear.extend(person[key][lang][2])
            deathyear.extend(person[key][lang][3])
            birthplace.append(person[key][lang][5])
            new_dict[key]["name"] = list(set(names))
            new_dict[key]["gender"] = list(set(gender))
            new_dict[key]["birthyear"] = list(set(birthyear))
            new_dict[key]["deathyear"] = list(set(deathyear))
            new_dict[key]["birthplace"] = list(set(birthplace))

    longest_match = {}
    for key in new_dict.keys():
        score_name = 0
        score_wikipedia = 0
        if key in name and key in wikipedia:
            if name[key][1]["Q-id"] == wikipedia[key][1]["Q-id"]:
                longest_match[key] = name[key]
            else:
                for k in name[key][1].keys():
                    if k == "Q-id":
                        continue
                    if new_dict[key][k]:
                        if k in ["birthyear", "deathyear"]:
                            if int(name[key][1][k]) in new_dict[key][k]:
                                score_name += 1
                        else:
                            if name[key][1][k] in new_dict[key][k]:
                                score_name += 1

                for k in wikipedia[key][1].keys():
                    if k == "Q-id":
                        continue
                    if new_dict[key][k]:
                        if k in ["birthyear", "deathyear"]:
                            if int(wikipedia[key][1][k]) in new_dict[key][k]:
                                score_wikipedia += 1
                        else:
                            if wikipedia[key][1][k] in new_dict[key][k]:
                                score_wikipedia += 1

                if score_name > score_wikipedia:
                    longest_match[key] = name[key]
                elif score_name < score_wikipedia:
                    longest_match[key] = wikipedia[key]
                else:  # same score
                    print("Check manually.")
        elif key in name:
            longest_match[key] = name[key]
        elif key in wikipedia:
            longest_match[key] = wikipedia[key]

    with open("../results/match_after_choosing_longest_match.json", "w") as f:
        json.dump(longest_match, f)

    #################################################################
    ################## Difference between the results of two approaches ##################

    difference_between_two_approaches = []
    l = [x for x in wikipedia.keys() if x in name.keys()]

    print("number of intersection: ", len(l))  # 374
    for key in l:
        if name[key] != wikipedia[key]:
            print("\nFor person with id ", key, ":")
            print("name[key]: ", "\n", name[key])
            print("wikipedia[key]: ", "\n", wikipedia[key], "\n")
            difference_between_two_approaches.append(key)



import logging
import os.path
import sys

import pandas as pd

from src.scripts.authenticity_space import read_space_csv

PERSON_GITHUB = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Person.csv"
SPACE_GITHUB = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Space.csv"
INST_GITHUB = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Institution.csv"
AGENT_GITHUB = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Agent.csv"

logger = logging.getLogger(__name__)


def search_wikidataID_in_table(row, agent_processed, agent_id):
    df_tmp = agent_processed[agent_processed["agent_id"] == row[agent_id]]
    # ToDo(QG): here should raise an error if the length > 1
    if len(df_tmp.index) == 1:
        if df_tmp.iloc[0]["wikidata_id"].startswith("Q"):
            row["wikidata_id"] = df_tmp.iloc[0]["wikidata_id"]
    return row


def addWikidataID_and_replaceSpace(
    df_PI_gh, agent_processed, agent_id, place_dict, entity_type
):
    for index, row in df_PI_gh.iterrows():
        if entity_type == "Institution":
            if row["place"] in place_dict:
                row["place"] = place_dict[row["place"]][0]
        elif entity_type == "Person":
            if row["place_of_birth"] in place_dict:
                row["place_of_birth"] = place_dict[row["place_of_birth"]][0]
        result = search_wikidataID_in_table(row, agent_processed, agent_id)
        df_PI_gh.loc[index] = result
    return df_PI_gh


def combine_agent_tables(df_agent_user, df_agent_gh, agent_ids_gh):
    """
    This function aims to use merging the Agent tables in ReadAct and in user's directory. If any agent_id in the
    user table already exists in ReadAct, delete according line(s) in the user table and then merge.
    :param df_agent_user: dataframe of the Agent.csv from user
    :param df_agent_gh: dataframe of the Agent.csv from ReadAct
    :param agent_ids_gh: all the agent_ids in the Agent.csv in ReadAct
    :return: a processed combined agent dataframe
    """
    df_agent_user_not_in_gh = df_agent_user.loc[
        ~df_agent_user["agent_id"].isin(agent_ids_gh)
    ]  # delete lines which
    # has agent_id in ReadAct, which is to say, for any agent_id already in ReadAct, use the ReadAct resource
    agent_cols = [
        "agent_id",
        "old_id",
        "agent_type",
        "wikidata_id",
        "fictionality",
        "language",
        "commentary",
        "note",
        "created",
        "created_by",
        "last_modified",
        "last_modified_by",
    ]
    df_processd = pd.merge(
        df_agent_gh, df_agent_user_not_in_gh, on=agent_cols, how="outer"
    )
    return df_processd


def process_agent_tables(entity_type, user_or_ReadAct, path):
    place_dict = read_space_csv()
    df_agent_gh = pd.read_csv(AGENT_GITHUB).fillna("")  # Get Agent table from ReadAct
    agent_ids_gh = list(
        set(df_agent_gh["agent_id"].tolist())
    )  # Get all the unique agent_ids
    last_item_id_gh = agent_ids_gh[-1]  # Get the last agent_id in ReadAct
    if entity_type == "Person":
        agent_id = "person_id"  # Set variables for later
        place_name = "place_of_birth"
    elif entity_type == "Institution":
        agent_id = "inst_id"
        place_name = "place"

    if user_or_ReadAct == "ReadAct":
        if entity_type == "Person":
            which_agent = PERSON_GITHUB
        elif entity_type == "Institution":
            which_agent = INST_GITHUB
        agent_processed = df_agent_gh
        dtype_dict = {}

    elif user_or_ReadAct == "user":
        if entity_type == "Person":
            which_agent = path[0]  # path of Person.csv in user's computer
            dtype_dict = {"birthyear": str, "deathyear": str}
        elif entity_type == "Institution":
            which_agent = path[0]  # path of Institution.csv in user's computer
            dtype_dict = {
                "start": str,
                "end": str,
                "alt_start": str,
                "alt_end": str,
            }  # here we assume all the years in Institution are type int
        df_agent_user = pd.read_csv(path[1]).fillna("")  # Get agent table

        # Check if agent_id are unique in user file
        if not pd.Series(df_agent_user["agent_id"]).is_unique:
            logger.error("Error: agent IDs in your Agent table are not unique.")
            sys.exit()

        agent_processed = combine_agent_tables(df_agent_user, df_agent_gh, agent_ids_gh)

    all_wikidata_ids = [x for x in agent_processed["wikidata_id"].tolist() if x]
    df_PI = pd.read_csv(which_agent, dtype=dtype_dict).fillna("")

    # Check if the place in user's Person/Institution table are all in ReadAct.
    # So that to get a processed space table to convert space IDs in P/I into space names.
    if all(place in place_dict for place in df_PI[place_name].tolist()):
        place_dict_combined = place_dict
        combined_two_space = False  # Unnecessary to check for potential local Space.csv
    else:  # place in user's table is not in ReadAct, check if any local space table
        space_path_user = path[1][:-9] + "Space.csv"
        if not os.path.isfile(space_path_user):
            logger.error(
                "There's place in your Institution/Person table without providing according space table. Please check."
            )
            sys.exit()
        else:
            combined_two_space = True  # Read local Space.csv
            place_dict_user = read_space_csv(space_path_user)
            # update place_dict by combining two Space table
            place_dict_combined = {
                **place_dict_user,
                **place_dict,
            }  # If any space_id in user's space table is also in ReadAct, take the value from ReadAct data

    df_PI["wikidata_id"] = ""  # add an empty wikidata_id column
    df_PI = addWikidataID_and_replaceSpace(
        df_PI, agent_processed, agent_id, place_dict_combined, entity_type
    )  # add wikidata_id information basing on ReadAct's Agent table

    return (
        df_PI,
        agent_processed,
        agent_ids_gh,
        last_item_id_gh,
        all_wikidata_ids,
        place_dict_combined,
        combined_two_space,
    )


if __name__ == "__main__":
    pass



"""
This is a python script to check authenticity of Named Entities of Space type in ReadAct.
Strategy:
- Read Space.csv in ReadAct, get space names and coordinates
- Search with coordinates in OpenStreetMap.
- If not match, use MediaWiki API service to look up with space name for QIDs
- Use SPARQL to retrieve properties from the found QIDs
- Compare wikidata item properties with data in the CSV table

The standards for macthing are:
1. With OpenStreetMap data, the lookup string should be contained in the location of the given coordinate.
2. With Wikidata, the difference between the coordinate in the CSV table and the coordinate in Wikidata should be
less or equal to 0.9.
"""

import json
import logging
import sys
import time
from itertools import islice

import pandas as pd
import requests

URL = "https://query.wikidata.org/sparql"
MEDIAWIKI_API_URL = "https://www.wikidata.org/w/api.php"
QUERY_COORDINATE = """
SELECT DISTINCT ?item ?coordinate
WHERE {{
  values ?item {{ wd:{} }}
  OPTIONAL {{ ?item  wdt:P625  ?coordinate . }}

  SERVICE wikibase:label {{ bd:serviceParam wikibase:language "en". }}
}}
"""

logger = logging.getLogger(__name__)


def read_space_csv(
    space_url="https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Space.csv",
):
    """
    A function to read "Space.csv" for now.
    :param space_url
    :return: a dictionary
    """
    df = pd.read_csv(space_url, error_bad_lines=False)
    geo_code_dict = {}
    for index, row in df.iterrows():
        # consider the case that if there are identical space_id in csv file
        if row["space_id"] not in geo_code_dict:
            # key: space_id
            # value: space_name, space_type, lat, long
            geo_code_dict[row["space_id"]] = [
                row["space_name"],
                row["space_type"],
                row["lat"],
                row["long"],
            ]
        else:
            logger.error("Space ID should be unique in this table. Please check. ")
            sys.exit()
    return geo_code_dict


def compare_to_openstreetmap(geo_code_dict):
    """
    A function to use the lat/lon from CSV file as lookup, to see if space_name is part of the returned message
    :param dictionary_value: tuple: space_name, lat, lon
    :return: None if match, a string message if not match
    """
    no_match_list = []
    for k, v in geo_code_dict.items():
        item = query_with_OSM(k, v)
        if item:  # item: space_name, space_type, lat, long, space_id
            no_match_list.append(item)
    return no_match_list


def query_with_OSM(k, v):
    if v[0] != "unknown" and v[2] != 0.0:
        lat = str(v[2])
        long = str(v[3])
        url = (
            "https://nominatim.openstreetmap.org/reverse?format=xml&lat="
            + lat
            + "&lon="
            + long
            + "&zoom=18&addressdetails=1&format=json&accept-language=en"
        )
        data = requests.get(url)
        if v[0].lower() not in str(data.json()).lower():
            item = v + [k]
            return item


def geo_code_compare(no_match_list):
    """
    For geo locations in Space.csv, compare latitude/longitude
    :param geo_code_dict
    :return: None or list of entries which don't match
    """
    still_no_match_list = []
    space_with_QID = {}  # To collect QIDs for Space.csv
    count = 0
    for i in no_match_list:  # i: space_name, space_type, lat, long, space_id
        if i[0] is None:
            res = None
        else:
            count += 1
            res = get_QID(
                i[0]
            )  # If there are more than one returned QID and we want to check all of them,
            # the following code must be modified as well.

        if count == 20:
            time.sleep(30)
            count = 0

        if res is None:
            still_no_match_list.append(i)
        else:
            coordinate_list = get_coordinate_from_wikidata(res["id"])
            # if no coordinate_list, collect item into list, break nested loop
            if len(coordinate_list) == 0:
                still_no_match_list.append(i)
                continue
            for coordinate_wiki in coordinate_list:
                # If the difference are within +-0.9, consider a match, no collection for no_match_list,
                # but one collect action for space_with_QID dictionary, then break nested loop
                # Pay attention that the query-returned coordinate have order: long, lat.
                if compare_coordinates_with_threhold(coordinate_wiki, i[2], i[3], 0.9):
                    space_with_QID[i[-1]] = i[:-1] + [res]
                    i = ""
                    break
            if len(i) > 0:
                still_no_match_list.append(i)
    print("space_with_QID", space_with_QID)
    return still_no_match_list, space_with_QID


def compare_coordinates_with_threhold(coordinate_wiki, lat_usr, long_usr, threshold):
    if (abs(float(coordinate_wiki[0]) - float(long_usr)) <= threshold) and (
        abs(float(coordinate_wiki[1]) - float(lat_usr)) <= threshold
    ):
        return True
    return False


def get_QID(lookup):
    params = {
        "action": "wbsearchentities",
        "language": "en",
        "search": lookup,
        "format": "json",
        "limit": 10,
    }
    reply = requests.get(MEDIAWIKI_API_URL, params=params)
    reply.raise_for_status()
    search_results = reply.json()
    results = []
    if search_results["success"] != 1:
        return None
    else:
        for i in search_results["search"]:
            results.append({"id": i["id"], "label": i["label"]})
    if len(results) == 0:
        return None
    else:
        # Note(QG): this can be easily extended into a longer list to increase the possibility of matching. Only
        # return the first one now due to efficiency.
        return results[0]


def get_coordinate_from_wikidata(q):
    """
    A function to extract coordinate location (if exists) of a wikidata entity
    :param qname: a wikidata id
    :return: a list with tuples, each tuple is a (lat, long) combination
    """
    coordinate_list = []
    headers = {"User-Agent": "wikidatalookup/1.0.0"}
    with requests.Session() as s:
        response = s.get(
            URL,
            params={
                "format": "json",
                "query": QUERY_COORDINATE.format(q),
            },
            headers=headers,
        )
        if response.status_code == 200:  # a successful response
            results = response.json().get("results", {}).get("bindings")
            if len(results) == 0:
                pass
            else:
                for r in results:
                    # If this entity is not recorded in this space_wiki dictionary yet:
                    if "coordinate" in r:
                        if "value" in r["coordinate"]:
                            c = r["coordinate"]["value"][6:-1].split()
                            # for example, '[114.158611111,22.278333333]'
                            coordinate_list.append(c)
    return coordinate_list


def chunks(it, size):
    it = iter(it)
    return iter(lambda: tuple(islice(it, size)), ())


if __name__ == "__main__":

    # # To compare the extracting coordinate location with the info in Space.csv
    # space_url = (
    #     "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Space.csv"
    # )
    # geo_code_dict = read_space_csv(space_url)
    #
    # # To filter CSV entries with comparing to openstreetmap first
    # no_match_list = compare_to_openstreetmap(geo_code_dict)

    # # To compare the rest with wikidata info
    # all_still_no_match_list = []
    # dictionary_list = []
    # for chunk in chunks(no_match_list, 30):  # the digit here controls the batch size
    #     if len(chunk) > 0:
    #         l, d = geo_code_compare(chunk)
    #         if l is not None:
    #             all_still_no_match_list += l
    #         dictionary_list += [d]
    #         print("\n I am taking a break XD \n")
    #         time.sleep(
    #             10
    #         )  # for every a few  entries, let this script take a break of 90 seconds
    # print("Finished the whole iteration")
    # print(all_still_no_match_list)
    # print("dictionary_list", dictionary_list)
    # match_for_space = {k: v for x in dictionary_list for k, v in dict(x).items()}
    # print(match_for_space)
    # with open("../results/match_for_space.json", "w") as f:
    #     json.dump(match_for_space, f)

    """
    The following is a one-time script to deactivate the compare_to_openstreetmap function to get as much Q-ids as
    we can.
    """

    # To compare the extracting coordinate location with the info in Space.csv
    space_url = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Space.csv"
    geo_code_dict = read_space_csv(space_url)

    # Assume 0 match from openstreetmap
    no_match_list = []
    for k, v in geo_code_dict.items():
        if v[0] != "unknown" and v[2] != 0.0:
            item = v + [k]
            no_match_list.append(item)

    # To compare the rest with wikidata info
    all_still_no_match_list = []
    dictionary_list = []
    for chunk in chunks(no_match_list, 30):  # the digit here controls the batch size
        if len(chunk) > 0:
            l, d = geo_code_compare(chunk)
            if l is not None:
                all_still_no_match_list += l
            dictionary_list += [d]

            print("\n I am taking a break XD \n")
            time.sleep(
                10
            )  # for every a few  entries, let this script take a break of 90 seconds

    print("Finished the whole iteration")
    print("all_still_no_match_list:", all_still_no_match_list)
    print("dictionary_list", dictionary_list)
    match_for_space = {k: v for x in dictionary_list for k, v in dict(x).items()}
    print(match_for_space)
    with open("../results/match_for_space_without_Openstreetmap.json", "w") as f:
        json.dump(match_for_space, f)



import logging
import re
import sys
from datetime import date

import pandas as pd

from src.scripts.agent_table_processing import process_agent_tables
from src.scripts.authenticity_institution import get_QID_inst, sparql_inst

logger = logging.getLogger(__name__)


def modify_note_lastModified_lastModifiedBy(row, message, today):
    row["note"] += " " + message
    row["last_modified"] = today
    row["last_modified_by"] = "ReadActor"
    return row


def process_Inst(df, entity_type):
    # Check if (inst_id, inst_name) pairs are unique in user file
    id_name_pairs = []
    for pair in zip(df["inst_id"], df["inst_name"]):
        pair_str = "({})".format(",".join(pair))
        id_name_pairs.append(pair_str)
    if len(id_name_pairs) > len(set(id_name_pairs)):
        logger.error(
            "Error: (inst_id, inst_name) in your Institution table are not unique."
        )
        sys.exit()

    # Process the local Agent table
    (
        df_inst_gh,
        agent_processed,
        inst_ids_gh,
        last_inst_id,
        all_wikidata_ids,
        _,
        _,
    ) = process_agent_tables(entity_type, "ReadAct", path=[])
    # Process local table row by row
    for index, row in df.iterrows():
        print("-------------\nFor row ", index + 2, " :")
        print(row.tolist())
        row, last_inst_id = check_each_row_Inst(
            index, row, df_inst_gh, inst_ids_gh, last_inst_id, all_wikidata_ids
        )
        # make the format of start and end (year) valid
        row = format_year_Inst(row)
        df.loc[index] = row
    return df


def check_each_row_Inst(
    index, row, df_inst_gh, inst_ids_gh, last_inst_id, all_wikidata_ids
):
    today = date.today().strftime("%Y-%m-%d")
    if row["note"] == "skip" or row["note"] == "Skip":
        return row, last_inst_id
    else:
        if isinstance(row["inst_id"], str) and len(row["inst_id"]) > 0:
            if row["inst_id"] in inst_ids_gh:  # inst_id is in ReadAct
                return (
                    __compare_wikidata_ids_Inst(index, row, df_inst_gh, today),
                    last_inst_id,
                )
            else:  # inst_id not in ReadAct
                if (isinstance(row["wikidata_id"], str) is True) and (
                    len(row["wikidata_id"]) > 0
                ):  # user did input wikidata_id
                    if (
                        row["wikidata_id"] in all_wikidata_ids
                    ):  # the input wikidata_id in ReadAct
                        logger.error(
                            "For row %s , wikidata_id in ReadAct but inst_id not. Please check. "
                            % index
                        )
                        sys.exit()
                    else:  # the input wikidata_id not in ReadAct
                        inst_wiki = sparql_inst(
                            [row["wikidata_id"]]
                        )  # query by wikidata_id to get other properties
                        l = [
                            inst_wiki["headquarters"],
                            inst_wiki["administrativeTerritorialEntity"],
                            inst_wiki["locationOfFormation"],
                        ]
                        l = [i for x in l for i in x]
                        if not any(l):  # all items in above list are empty strings
                            message = (
                                "For row %s, the user input wikidata_id does not have relevant info. Couldn't check. "
                            ) % index
                            logger.warning(message)
                            row = modify_note_lastModified_lastModifiedBy(
                                row, message, today
                            )
                        else:
                            row = __compare_place_and_start_for_Inst(
                                index, row, inst_wiki, l, today, row["wikidata_id"]
                            )
                else:  # user did NOT input wikidata_id
                    # TODO(QG): maybe it is better to modify get_QID_inst()
                    if row["inst_name"] is None or len(row["inst_name"]) == 0:
                        wikidata_id_from_query_Inst = None
                    else:
                        wikidata_id_from_query_Inst = get_QID_inst(
                            row["inst_name"]
                        )  # only return one value
                    if (
                        wikidata_id_from_query_Inst is None
                    ):  # query by name and return None
                        logger.info(
                            "For row %s : Query by name didn't find any wikidata item."
                            % index
                        )
                    else:  # query by name and find a wikidata_id
                        if (
                            wikidata_id_from_query_Inst[0]["id"] in all_wikidata_ids
                        ):  # found wikidata_id in ReadAct
                            logger.error(
                                'For row %s, The wikidata_id found by querying with institution name exists in ReadAct already. If you are 100% sure that your input is correct, you can put "skip" in "note". '
                                % index
                            )
                            sys.exit()
                        # The found wikidata_id is not in ReadAct, the next step is to check its coordinate
                        else:
                            inst_wiki = sparql_inst(
                                [wikidata_id_from_query_Inst[0]["id"]]
                            )  # query by wikidata_id to get other properties
                            l = [
                                inst_wiki["headquarters"],
                                inst_wiki["administrativeTerritorialEntity"],
                                inst_wiki["locationOfFormation"],
                            ]
                            l = [i for x in l for i in x]
                            if not any(l):  # all items in above list are empty strings
                                message = (
                                    "For row %s, the user input wikidata_id does not have relevant info. Couldn't check. "
                                ) % index
                                logger.warning(message)
                                row = modify_note_lastModified_lastModifiedBy(
                                    row, message, today
                                )
                            else:
                                row = __compare_place_and_start_for_Inst(
                                    index,
                                    row,
                                    inst_wiki,
                                    l,
                                    today,
                                    wikidata_id_from_query_Inst[0]["id"],
                                )
        else:  # user did not input inst_id
            logger.error("Please input inst_id for row %s ." % index)
            sys.exit()
    return row, last_inst_id


def __compare_wikidata_ids_Inst(index, row, df_inst_gh, today):
    wikidata_id_usr = row["wikidata_id"]
    row_gh_index = df_inst_gh.index[
        (df_inst_gh["inst_id"] == row["inst_id"])
        & (df_inst_gh["language"] == row["language"])
    ].tolist()[0]
    row_GH = df_inst_gh.iloc[row_gh_index]
    wikidata_id_gh = row_GH["wikidata_id"]
    if wikidata_id_gh == wikidata_id_usr:  # two wikidata_id are the same
        res = __compare_two_rows_Inst(row, row_GH)
        if not res:  # one or multiple fields don't match
            return __overwrite_Inst(row, row_GH, index, today)
        logger.info("Row %s is checked. Pass " % index)  # all the fields match
        return row
    else:  # two wikidata_id are not the same
        error_msg = (
            "For row %s : `wikidata_id` does not match GitHub data. Please check. "
            % index
        )
        logger.error(error_msg)
        sys.exit()


def __compare_two_rows_Inst(row, row_gh):
    fields_to_be_compared = [
        "inst_name",
        "language",
        "place",
        "start",
        "end",
        "alt_start",
        "alt_end",
        "inst_alt_name",
        "wikidata_id",
        "note",
        "source",
        "page",
        "created",
        "created_by",
        "last_modified",
        "last_modified_by",
    ]
    for i in fields_to_be_compared:
        if row_gh[i] == "":
            continue
        elif i in ["start", "end", "alt_start", "alt_end"]:
            if "." in str(row[i]):
                row[i] = int(
                    float(row[i])
                )  # to remove the ".0" part of any potential floating-point number
            if bool(re.match("[0-9.]+", str(row[i]))):
                if int(row[i]) != int(row_gh[i]):
                    return False
        else:
            if row[i] != row_gh[i]:
                return False
    return True


def __overwrite_Inst(row, row_gh, index, today):
    fields_to_be_overwritten = [
        "inst_name",
        "language",
        "place",
        "start",
        "end",
        "alt_start",
        "alt_end",
        "inst_alt_name",
        "wikidata_id",
        "note",
        "source",
        "page",
        "created",
        "created_by",
        "last_modified",
        "last_modified_by",
    ]
    modified_fields = []
    for i in fields_to_be_overwritten:
        if row[i] != row_gh[i]:
            row[i] = row_gh[i]
            modified_fields.append(i)
    message = "In row %s , the following fields are overwritten: %s " % (
        index,
        ", ".join(map(str, modified_fields)),
    )
    row = modify_note_lastModified_lastModifiedBy(row, message, today)
    logger.info(message)
    return row


def __compare_place_and_start_for_Inst(index, row, inst_wiki, l, today, wikidata_id):
    if (row["place"] in l) or (
        str(row["start"])[0:4] in str(inst_wiki["inception"])[0:4]
    ):
        if wikidata_id != row["wikidata_id"]:
            row["wikidata_id"] = wikidata_id
            message = (
                "For row %s, the wikidata_id can be queried. You should look institution up in Wikidata and write the wikidata_id in your table in the future. "
                % index
            )
            logger.warning(message)
            row = modify_note_lastModified_lastModifiedBy(row, message, today)
        else:
            logger.info("Row %s is checked. Pass. " % index)
    else:
        message = (
            "Row %s has no field which is matching any according wikidata properties. Please check. "
            % index
        )
        logger.warning(message)
        row = modify_note_lastModified_lastModifiedBy(row, message, today)
    return row


def format_year_Inst(row):
    for x in ["start", "end"]:
        if isinstance(row[x], float):
            row[x] = int(row[x])
        if "-" in str(row[x]):
            if isinstance(row[x][1:], float):
                row[x] = "-" + str(int(row[x][1:]))
            if isinstance(row[x][1:], int):
                year = str(int(row[x][1:]))
                if len(year) < 3:
                    pad = 3 - len(year)  # pad = 2,1
                    if pad == 1:
                        row["start"] = "-" + "0" + year
                    elif pad == 2:
                        row["start"] = "-" + "00" + year
        elif isinstance(row[x], int):
            year = str(int(row[x]))
            if len(year) < 4:
                pad = 4 - len(year)  # pad = 3,2,1
                if pad == 1:
                    row[x] = "0" + year
                elif pad == 2:
                    row[x] = "00" + year
                elif pad == 3:
                    row[x] = "000" + year
    return row


if __name__ == "__main__":
    pass



"""
This is a python script to check authenticity of Named Entities of Institution type in ReadAct.
Strategy:
- Read Institution.csv in ReadAct, get lookups
- Get the QIDs by using MediaWiki API service
- Use SPARQL to retrieve properties from the found QIDs
- Compare wikidata item properties with data in the CSV table

The standards of verificating an Institution entry is:
to find one match among headquarters/administrativeTerritorialEntity/locationOfFormation/inception.
"""
import json
import sys
import time

import pandas as pd
import requests

from src.scripts.authenticity_space import read_space_csv

MEDIAWIKI_API_URL = "https://www.wikidata.org/w/api.php"
URL = "https://query.wikidata.org/sparql"
QUERY1 = """
PREFIX  schema: <http://schema.org/>
PREFIX  bd:   <http://www.bigdata.com/rdf#>
PREFIX  wdt:  <http://www.wikidata.org/prop/direct/>
PREFIX  wikibase: <http://wikiba.se/ontology#>

SELECT DISTINCT  ?item ?itemLabel ?headquartersLabel ?administrativeTerritorialEntityLabel 
?locationOfFormationLabel ?inceptionLabel
WHERE
  {{ ?article schema:about ?item ;
    FILTER ( ?item = <http://www.wikidata.org/entity/"""

QUERY2 = """> )
    OPTIONAL
      { ?item  wdt:P159  ?headquarters }
    OPTIONAL
      { ?item  wdt:P131  ?administrativeTerritorialEntity }
    OPTIONAL
      { ?item wdt:P740   ?locationOfFormation }
    OPTIONAL
      { ?item wdt:P571  ?inception }
    SERVICE wikibase:label
      { bd:serviceParam wikibase:language  "[AUTO_LANGUAGE], en"
      }
  }}
GROUP BY ?item ?itemLabel ?headquartersLabel ?administrativeTerritorialEntityLabel 
?locationOfFormationLabel  ?inceptionLabel 
"""


def read_institution_csv(
    inst_url="https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Institution.csv",
):
    """
    A function to read "Institution.csv"
    :param inst_url: the GitHub address of Institution.csv
    :return: a dictionary
    """
    df = pd.read_csv(inst_url, error_bad_lines=False)
    df = df.fillna("")
    ins_dict = {}
    place_dict = read_space_csv()
    for index, row in df.iterrows():
        if (row["inst_id"], row["inst_name"]) not in ins_dict:
            if row["place"] in place_dict:
                if len(str(row["start"])) > 0:
                    row["start"] = int(float(row["start"]))
                if len(str(row["end"])) > 0:
                    row["end"] = int(float(row["end"]))
                ins_dict[(row["inst_id"], row["inst_name"])] = [
                    place_dict[row["place"]][0],
                    row["start"],
                    row["end"],
                ]
            else:
                ins_dict[(row["inst_id"], row["inst_name"])] = [
                    row["place"],
                    row["start"],
                    row["end"],
                ]
                print("Please check. A space_id is not in Space.csv.")
        else:
            print(
                "Please check. The combination of inst_id and inst_name should be unique."
            )
            sys.exit()
    return ins_dict


def compare_inst(inst_dict, sleep=2):
    no_match = {}
    match = {}
    for k, v in inst_dict.items():
        results = get_QID_inst(k[1])
        if results is None:
            no_match[k] = v
            continue
        q_ids = [x["id"] for x in results if x is not None]
        if q_ids is None:
            no_match[k] = v
            continue
        inst_wiki_dict = sparql_inst(q_ids, sleep)
        if (
            len(inst_wiki_dict["headquarters"]) > 0
            and v[0] in inst_wiki_dict["headquarters"]
        ):
            match[k] = v + q_ids
            print("A match: ", k, v + q_ids)
        elif (
            len(inst_wiki_dict["administrativeTerritorialEntity"]) > 0
            and v[0] in inst_wiki_dict["administrativeTerritorialEntity"]
        ):
            match[k] = v + q_ids
            print("A match: ", k, v + q_ids)
        elif (
            len(inst_wiki_dict["locationOfFormation"]) > 0
            and v[0] in inst_wiki_dict["locationOfFormation"]
        ):
            match[k] = v + q_ids
            print("A match: ", k, v + q_ids)
        elif (
            len(inst_wiki_dict["inception"]) > 0
            and str(v[1])[0:4] in str(inst_wiki_dict["inception"])[0:4]
        ):
            match[k] = v + q_ids
            print("A match: ", k, v + q_ids)
        else:
            no_match[k] = v
    return no_match, match


def sparql_inst(q_ids, sleep=2):
    if q_ids is None:
        return []
    inst_wiki = {}
    with requests.Session() as s:
        for q in q_ids:
            response = s.get(
                URL, params={"format": "json", "query": QUERY1 + q + QUERY2}
            )
            if response.status_code == 200:  # a successful response
                results = response.json().get("results", {}).get("bindings")
                (
                    inst_wiki["name"],
                    inst_wiki["headquarters"],
                    inst_wiki["administrativeTerritorialEntity"],
                    inst_wiki["locationOfFormation"],
                    inst_wiki["inception"],
                    inst_wiki["QID"],
                ) = ([], [], [], [], [], [])
                if q is not None:
                    inst_wiki["QID"] = q
                if results:
                    for b in results:
                        if "itemLabel" in b:
                            inst_wiki["name"] = b["itemLabel"]["value"]
                        if "headquartersLabel" in b:
                            headquarters = b["headquartersLabel"]["value"]
                            inst_wiki["headquarters"].append(headquarters)
                        if "administrativeTerritorialEntityLabel" in b:
                            administrativeTerritorialEntity = b[
                                "administrativeTerritorialEntityLabel"
                            ]["value"]
                            inst_wiki["administrativeTerritorialEntity"].append(
                                administrativeTerritorialEntity
                            )
                        if "locationOfFormationLabel" in b:
                            locationOfFormation = b["locationOfFormationLabel"]["value"]
                            inst_wiki["locationOfFormation"].append(locationOfFormation)
                        if "inceptionLabel" in b:
                            inception = b["inceptionLabel"]["value"]
                            inst_wiki["inception"].append(inception)
            time.sleep(sleep)
    return inst_wiki


def get_QID_inst(lookup):
    params = {
        "action": "wbsearchentities",
        "language": "en",
        "search": lookup,
        "format": "json",
        "limit": 10,
    }
    reply = requests.get(MEDIAWIKI_API_URL, params=params)
    reply.raise_for_status()
    search_results = reply.json()

    results = []
    if search_results["success"] != 1:
        return None
    else:
        for i in search_results["search"]:
            results.append({"id": i["id"], "label": i["label"]})
    if len(results) == 0:
        return None
    else:
        return results[
            0:1
        ]  # Return the first Qid. In the future, to increase the accuracy of match, just rewrite
        # this line to allow more QIDs to be returned for comparing.


if __name__ == "__main__":

    inst_dict = read_institution_csv()
    print(inst_dict)
    no_match, match = compare_inst(inst_dict, 10)
    print("no_match dictionary: ", no_match)
    print("length of the no_match dictionary: ", len(no_match))

    print("=======\n\n\n\n\n")
    print("match dictionary: ", match)
    print("length of the match dictionary: ", len(match))

    d = {}
    for k, v in match.items():
        d[k[0]] = [k[1]] + v
    with open("../results/match_for_institution.json", "w") as f:
        json.dump(d, f)



import unittest
from datetime import date

import pandas as pd

from src.scripts.agent_table_processing import process_agent_tables
from src.scripts.process_Person import check_each_row_Person


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.column_names = [
            "person_id",
            "family_name",
            "first_name",
            "language",
            "sex",
            "birthyear",
            "deathyear",
            "place_of_birth",
            "wikidata_id",
            "created",
            "created_by",
            "last_modified",
            "last_modified_by",
            "note",
        ]
        self.l = ["", "", "", "", "", "", "", "", "", "", "", "", "", ""]
        self.df = pd.DataFrame([self.l])
        self.df.columns = self.column_names
        self.df = self.df.astype(str)

        (
            self.df_person_new,
            _,
            self.person_ids_gh,
            self.last_person_id,
            self.wikidata_ids_GH,
            _,
            _,
        ) = process_agent_tables("Person", "ReadAct", [])
        self.today = date.today().strftime("%Y-%m-%d")

    def test_it_should_respect_skip_annotation(self):
        self.l = [
            "AG0001",
            "鲁",
            "迅",
            "zh",
            "male",
            "0000",
            "0000",
            "Shaoxing",
            "Q23114",
            "2021-12-22",
            "QG",
            "",
            "",
            "skip",
        ]
        self.df.loc[0] = self.l
        self.assertEqual(
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )[0].tolist(),
            [
                "AG0001",
                "鲁",
                "迅",
                "zh",
                "male",
                "0000",
                "0000",
                "Shaoxing",
                "Q23114",
                "2021-12-22",
                "QG",
                "",
                "",
                "skip",
            ],
        )

    def test_it_should_NOT_skip_without_annotation(self):
        self.l = [
            "AG0001",
            "鲁",
            "迅",
            "zh",
            "male",
            "0000",
            "1111",
            "Shaoxing",
            "Q23114",
            "2021-12-22",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        self.assertNotEqual(
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )[0].tolist()[0:-1],
            [
                "AG0001",
                "鲁",
                "迅",
                "zh",
                "male",
                "1881",
                "1936",
                "Shaoxing",
                "Q23114",
                "2021-12-22",
                "QG",
                self.today,
                "ReadActor",
            ],
        )

    # def test_it_should_match_all_witout_change(self):
    #     self.l = [
    #         "AG0001",
    #         "鲁",
    #         "迅",
    #         "zh",
    #         "male",
    #         "1881",
    #         "1936",
    #         "SP0048",
    #         "Q23114",
    #         "2017-07-03",
    #         "LH",
    #         "2022-05-03",
    #         "ReadActor",
    #         "",
    #     ]
    #     self.df.loc[0] = self.l
    #     self.assertEqual(
    #         check_each_row_Person(
    #             0,
    #             self.df.iloc[0],
    #             self.df_person_new,
    #             self.person_ids_gh,
    #             self.last_person_id,
    #             self.wikidata_ids_GH,
    #         )[0].tolist()[0:-1],
    #         [
    #             "AG0001",
    #             "鲁",
    #             "迅",
    #             "zh",
    #             "male",
    #             "1881",
    #             "1936",
    #             "SP0048",
    #             "Q23114",
    #             "2017-07-03",
    #             "LH",
    #             "2022-05-03",
    #             "ReadActor",
    #         ],
    #     )

    # def test_it_should_update_data_for_full_id_match(self):
    #     self.l = [
    #         "AG0001",
    #         "鲁",
    #         "迅",
    #         "zh",
    #         "male",
    #         "1881",
    #         "1936",
    #         "Shaoxing",
    #         "Q23114",
    #         "2021-12-22",
    #         "QG",
    #         "",
    #         "",
    #         "",
    #     ]
    #     self.df.loc[0] = self.l
    #     self.assertEqual(
    #         check_each_row_Person(
    #             0,
    #             self.df.iloc[0],
    #             self.df_person_new,
    #             self.person_ids_gh,
    #             self.last_person_id,
    #             self.wikidata_ids_GH,
    #         )[0].tolist()[0:-1],
    #         [
    #             "AG0001",
    #             "鲁",
    #             "迅",
    #             "zh",
    #             "male",
    #             "1881",
    #             "1936",
    #             "SP0048",
    #             "Q23114",
    #             "2017-07-03",
    #             "LH",
    #             "2022-05-03",
    #             "ReadActor",
    #         ],
    #     )

    def test_it_should_fail_on_ID_conflict(self):
        self.l = [
            "AG0001",
            "鲁",
            "迅",
            "zh",
            "male",
            "1881",
            "1936",
            "Shaoxing",
            "Q00000000",
            "2021-12-22",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        with self.assertRaises(SystemExit) as cm:
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )
        self.assertEqual(cm.exception.code, None)

    def test_it_should_not_reuse_wikiID(self):
        self.l = [
            "AG1200",
            "Monet",
            "Claude",
            "en",
            "male",
            "1840",
            "1926",
            "Paris",
            "Q23114",
            "2021-12-22",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        with self.assertRaises(SystemExit) as cm:
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )
        self.assertEqual(cm.exception.code, None)

    # TODO(DP): I do not understand how this testcase is different from the second test above
    # (QG): This should be a new entry which does not exist in ReadAct. The second test above is for an existed person.
    def test_it_should_not_update_new_entries_with_matching_data(self):
        self.l = [
            "AG0000",
            "Musk",
            "Elon",
            "en",
            "male",
            "1971",
            "",
            "Pretoria",
            "Q317521",
            "2022-06-07",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        self.assertEqual(
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )[0].tolist()[0:-1],
            [
                "AG0000",
                "Musk",
                "Elon",
                "en",
                "male",
                "1971",
                "",
                "Pretoria",
                "Q317521",
                "2022-06-07",
                "QG",
                "",
                "",
            ],
        )

    def test_it_should_update_new_entries_with_missing_data(self):
        self.l = [
            "AG0000",
            "Musk",
            "Elon",
            "en",
            "",
            "",
            "",
            "",
            "Q317521",
            "2022-06-07",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        self.assertEqual(
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )[0].tolist()[0:-1],
            [
                "AG0000",
                "Musk",
                "Elon",
                "en",
                "male",
                "1971",
                "",
                "Pretoria",
                "Q317521",
                "2022-06-07",
                "QG",
                self.today,
                "ReadActor",
            ],
        )

    def test_it_should_use_wikiID_to_dedupe_readActID(self):
        self.l = [
            "AG1200",
            "鲁",
            "迅",
            "zh",
            "",
            "",
            "",
            "",
            "",
            "2021-12-22",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        with self.assertRaises(SystemExit) as cm:
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )
        self.assertEqual(cm.exception.code, None)

    def test_it_should_update_wikiID_using_persondata(self):
        self.l = [
            "AG0000",
            "Musk",
            "Elon",
            "en",
            "male",
            "1971",
            "",
            "Pretoria",
            "",
            "2022-06-07",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        self.assertEqual(
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )[0].tolist()[0:-1],
            [
                "AG0000",
                "Musk",
                "Elon",
                "en",
                "male",
                "1971",
                "",
                "Pretoria",
                "Q317521",
                "2022-06-07",
                "QG",
                self.today,
                "ReadActor",
            ],
        )

    def test_it_should_update_data_with_missing_wikiID(self):
        self.l = [
            "AG0000",
            "Musk",
            "Elon",
            "en",
            "male",
            "1971",
            "",
            "Tokyo",
            "",
            "2022-06-07",
            "QG",
            "",
            "",
            "",
        ]
        self.df.loc[0] = self.l
        self.assertEqual(
            check_each_row_Person(
                0,
                self.df.iloc[0],
                self.df_person_new,
                self.person_ids_gh,
                self.last_person_id,
                self.wikidata_ids_GH,
            )[0].tolist()[0:-1],
            [
                "AG0000",
                "Musk",
                "Elon",
                "en",
                "male",
                "1971",
                "",
                "Pretoria",
                "Q317521",
                "2022-06-07",
                "QG",
                self.today,
                "ReadActor",
            ],
        )
        # should test the warning and clean up after testing updated files


if __name__ == "__main__":
    unittest.main()



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="ReadActor",
    version="1.0.1",
    author="Qin Gu",
    author_email="guqin7@gmail.com",
    description="A lookup tool for ReadChina project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/readchina/ReadActor",
    project_urls={
        "Bug Tracker": "https://github.com/pypa/ReadActor/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "Click",
    ],
    entry_points={
        "console_scripts": [
            "readactor = src.scripts.readactor:cli",
        ],
    },
)



import logging
import sys
from datetime import date

import pandas as pd

from src.scripts.authenticity_space import (
    compare_coordinates_with_threhold,
    get_coordinate_from_wikidata,
    get_QID,
    query_with_OSM,
)

SPACE_GITHUB = "https://raw.githubusercontent.com/readchina/ReadAct/2.0-RC-patch/csv/data/Space.csv"

logger = logging.getLogger(__name__)


def process_Spac(df):
    # Check if space_id are unique in user file
    if not pd.Series(df["space_id"]).is_unique:
        logger.error("Error: space IDs in your Space table are not unique.")
        sys.exit()

    # Read the Space table in ReadAct
    df_space_gh = pd.read_csv(SPACE_GITHUB)
    df_space_gh = df_space_gh.fillna("")  # Replace all the nan into empty string
    check_gh(df_space_gh)
    space_ids_gh = df_space_gh["space_id"].tolist()
    space_ids_gh.sort()
    wikidata_ids_GH = df_space_gh["wikidata_id"].tolist()
    last_space_id = space_ids_gh[-1]
    # Process local table row by row
    for index, row in df.iterrows():
        print(
            "-------------\nFor row ", index + 2, " :"
        )  # Because the header line in Person.csv is already row 1
        print(row.tolist())
        row, last_space_id = check_each_row_Space(
            index, row, df_space_gh, space_ids_gh, last_space_id, wikidata_ids_GH
        )
        df.loc[index] = row
    return df


def check_each_row_Space(
    index, row, df_space_gh, space_ids_gh, last_space_id, wikidata_ids_GH
):
    today = date.today().strftime("%Y-%m-%d")
    if (
        row["note"].strip() == "skip" or row["note"].strip() == "Skip"
    ):  # user wants to skip this line
        logger.info("User chooses to skip row %s ." % index)
    elif row["note"].strip() != "skip" or row["note"].strip() != "Skip":
        if (
            isinstance(row["space_id"], str) and len(row["space_id"]) > 0
        ):  # user did input space_id
            if row["space_id"] in space_ids_gh:  # space_id in ReadAct
                return (
                    __compare_wikidata_ids_Space(index, row, df_space_gh, today),
                    last_space_id,
                )
            else:  # space_id not in ReadAct
                if (isinstance(row["wikidata_id"], str) is True) and (
                    len(row["wikidata_id"]) > 0
                ):  # user did input wikidata_id
                    if (
                        row["wikidata_id"] in wikidata_ids_GH
                    ):  # the input wikidata_id in ReadAct
                        logger.error(
                            "For row %s , wikidata_id in ReadAct but space_id not. Please check."
                            % index
                        )
                        sys.exit()
                    else:  # the input wikidata_id not in ReadAct
                        coordinate_from_wikidata = get_coordinate_from_wikidata(
                            row["wikidata_id"]
                        )
                        if (
                            len(coordinate_from_wikidata) == 0
                        ):  # input wikidata_id has no P625
                            logger.info(
                                "In row %s, the user input wikidata_id does not have coordinate location property ("
                                "P625). Could not check 'lat' and 'long'. Pass " % index
                            )
                        else:  # input wikidata_id has P625 (coordinate location)
                            # TODO(QG): the threshold here is +/- 0.1, should discuss if it makes sense.
                            if compare_coordinates_with_threhold(
                                coordinate_from_wikidata, row["lat"], row["long"], 0.1
                            ):
                                logger.info("Row %s is checked. Pass." % index)
                            else:  # if the difference between coordinates bigger than threshold
                                warning_msg = (
                                    "In row %s ,you'd better compare the coordinate you entered and the one on "
                                    "Wikidata manually." % index
                                )
                                logger.warning("warning_msg")
                                row = modify_note_lastModified_lastModifiedBy(
                                    row, warning_msg, today
                                )
                else:  # user input space_id not in ReadAct and user did not input wikidata_id
                    res_OSM = query_with_OSM(
                        row["space_id"],
                        [row["space_name"], row["space_type"], row["lat"], row["long"]],
                    )
                    # TODO(QG): here, this subsection should be changed in the future to obtain the potential
                    #  wikidata id.
                    if (
                        res_OSM is None
                    ):  # space_name matches with geo coordinates basing on OSM query
                        logger.info(
                            "In row %s , space_name matches with geo coordinates according to OSM query."
                            % index
                        )
                    else:  # res_OSM: space_name, space_type, lat, long, space_id
                        # ToDo(QG): no_match_list[0][0] should be equal to row['space_name]. Should check.
                        wikidata_id_from_query = get_QID(
                            res_OSM[0]
                        )  # only return one value
                        if wikidata_id_from_query is None:  # query returns None
                            logger.info(
                                "For row %s : check lat+long on OSM found the address which doesn't contain "
                                "space_name. Query by name didn't find any wikidata item."
                                % index
                            )
                        else:  # query by name and find a wikidata_id
                            if (
                                wikidata_id_from_query in wikidata_ids_GH
                            ):  # found wikidata_id in ReadAct
                                logger.error(
                                    "For row %s, found wikidata_id already in ReadAct while the space_id is not. If "
                                    "you are certain about your input, you can put the word 'skip' in 'note' to avoid "
                                    "this error message. " % index
                                )
                                sys.exit()
                            # The found wikidata_id is not in ReadAct, the next step is to check its coordinate
                            else:
                                # ToDo(QG): for computing efficiency, here the `no_match_list` should be replaced
                                #  with `wikidata_id_query` but the code should be modified accordingly
                                coordinate_from_query = get_coordinate_from_wikidata(
                                    wikidata_id_from_query
                                )
                                if (
                                    len(coordinate_from_query) == 0
                                ):  # has no geo location property
                                    logger.info(
                                        "In row %s, the user input wikidata_id does not have coordinate location "
                                        "property (P625). Could not check 'lat' and 'long'. "
                                        % index
                                    )
                                else:  # compare user input geo coordinate with the one from query
                                    # TODO(QG): the threshold here is +/- 0.1, should discuss if it makes sense.
                                    if compare_coordinates_with_threhold(
                                        coordinate_from_query,
                                        row["lat"],
                                        row["long"],
                                        0.1,
                                    ):  # consider a match
                                        row[
                                            "wikidata_id"
                                        ] = wikidata_id_from_query  # to update wikidata
                                        warning_msg = (
                                            "For cases like row %s , you'd better look the Space up in wikidata "
                                            "and input wikidata_id in your table."
                                            % index
                                        )
                                        row = modify_note_lastModified_lastModifiedBy(
                                            row, warning_msg, today
                                        )
                                        logger.warning(warning_msg)
                                    else:  # geo coordinates conflict. Probably the wikidata query is not accurate.
                                        # Or the query returns a geo location in the same area but different from the
                                        # geo location given by user since a geo area has a certain size.
                                        warning_msg = (
                                            "For row %s, by querying with space_name, the found wikidata item "
                                            "has a different geo location from your input. You are suggested to "
                                            "check "
                                            "it and put the word 'skip' in 'note' if you are confident about "
                                            "your input." % index
                                        )
                                        row = modify_note_lastModified_lastModifiedBy(
                                            row, warning_msg, today
                                        )
                                        logger.warning(warning_msg)
        else:  # user did not input space_id
            logger.error("Please input space_id for row %s ." % index)
            sys.exit()
    return row, last_space_id


def modify_note_lastModified_lastModifiedBy(row, message, today):
    row["note"] += " " + message
    row["last_modified"] = today
    row["last_modified_by"] = "ReadActor"
    return row


def __compare_wikidata_ids_Space(index, row, df_space_gh, today):
    """
    When user input wikidata_id and this wikidata_id already exists in ReadAct,
    compare the input row with the ReadAct row which has the same "sapce_id":
    1. compare the wikidata ids.
    2. if both wikidata ids are identical, then compare the rest fields. Otherwise, use ReadAct data to rewrite the
    user input.
    3. if two wikidata ids are not identical, report error for mismatch.
    """
    wikidata_id_usr = row["wikidata_id"]
    row_gh_index = df_space_gh.index[
        (df_space_gh["space_id"] == row["space_id"])
        & (df_space_gh["language"] == row["language"])
    ].tolist()[0]
    row_GH = df_space_gh.iloc[row_gh_index]
    wikidata_id_gh = row_GH["wikidata_id"]
    if (
        wikidata_id_gh is None or len(wikidata_id_gh) == 0
    ):  # if there is no Wikidata_id in ReadAct, pass.
        return row
    if wikidata_id_gh == wikidata_id_usr:
        res = __compare_two_rows_Space(row, row_GH)
        if not res:
            return __overwrite_Space(row, row_GH, index, today)
        logger.info("Row %s is checked. Pass." % index)
        return row
    else:
        logger.error(
            "In row %s , compare with the same space_id in ReadAct, you input a conflicting wikidata_id. "
            "Please check." % index
        )
        sys.exit()


def __compare_two_rows_Space(row, row_GH):
    fields_to_be_compared = ["space_type", "space_name", "lat", "long"]
    for i in fields_to_be_compared:
        if row[i] != row_GH[i]:
            return False
    return True


def __overwrite_Space(row, row_gh, index, today):
    fields_to_be_overwritten = [
        "space_type",
        "space_name",
        "language",
        "lat",
        "long" "created",
        "created_by",
        "last_modified",
        "last_modified_by",
        "note",
    ]
    modified_fields = []
    for i in fields_to_be_overwritten:
        if row[i] != row_gh[i]:
            row[i] = row_gh[i]
            modified_fields.append(i)
    message = "In row %s , the following fields are overwritten: %s " % (
        index,
        ", ".join(map(str, modified_fields)),
    )
    row = modify_note_lastModified_lastModifiedBy(row, message, today)
    logger.info(message)
    return row


def check_gh(
    df,
):  # a function to check if Person.csv on GitHub has `wikidata_id` column
    if "wikidata_id" not in df.columns:
        error_msg = (
            "There is no `wikidata_id` column in the accordingly CSV table on GitHub. Please inform someone to check "
            "it. "
            ""
            "By ReadActor."
        )
        logger.error(error_msg)
        exit()


if __name__ == "__main__":
    pass
